""" Represents the data type class. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: data_type_class_base.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

from abc import abstractmethod
from typing import Any
from pasta_eln.GUI.dataverse.data_type_class_context import DataTypeClassContext


class DataTypeClass:
  """
  Represents the base data type class.
  """

  def __init__(self, context: DataTypeClassContext) -> None:
    """
    Initializes a new instance of the DataTypeClass class.

    Explanation:
        This method initializes a new instance of the DataTypeClass class.

    Args:
      context (DataTypeClassContext): The context of the data type class.

    Raises:
      TypeError: If the provided context is not of type DataTypeClassContext.
    """
    if not isinstance(context, DataTypeClassContext):
      raise TypeError('context must be of type DataTypeClassContext')
    self.context: DataTypeClassContext = context

  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Creates a new instance of the DataTypeClass class.

    Explanation:
        This method creates a new instance of the DataTypeClass class.

    Args:
        *_: Variable length argument list.
        **__: Arbitrary keyword arguments.

    Returns:
        Any: The new instance of the DataTypeClass class.

    """
    return super().__new__(cls)

  @abstractmethod
  def add_new_entry(self) -> None:
    """
    Saves the modifications made to the data type class.

    Explanation:
        This method saves the modifications made to the data type class.

    """
    return

  @abstractmethod
  def populate_entry(self) -> None:
    """
    Saves the modifications made to the data type class.

    Explanation:
        This method saves the modifications made to the data type class.

    """
    return

  @abstractmethod
  def save_modifications(self) -> None:
    """
    Saves the modifications made to the data type class.

    Explanation:
        This method saves the modifications made to the data type class.

    """
    return



""" dialog to edit docType schema """
from enum import Enum
from typing import Any
import numpy as np
import pandas as pd
from PySide6.QtWidgets import (QComboBox, QDialog, QDialogButtonBox, QInputDialog,  # pylint: disable=no-name-in-module
                               QMessageBox, QTabBar, QTableWidget, QTableWidgetItem, QTabWidget, QVBoxLayout)
from ...guiCommunicate import Communicate
from ...guiStyle import IconButton, Label, TextButton, showMessage, widgetAndLayout
from .delete_column_delegate import DeleteColumnDelegate
from .docTypeEdit import DocTypeEditor
from .mandatory_column_delegate import MandatoryColumnDelegate
from .name_column_delegate import NameColumnDelegate
from .reorder_column_delegate import ReorderColumnDelegate

#                0       1            2      3           4      5         6
COLUMN_NAMES = ['name','description','unit','mandatory','list','move up','delete']
COLUMN_WIDTH = [100,   250,          100,   80,         245,   60,       50]
pd.options.mode.copy_on_write = True

class SchemeEditor(QDialog):
  """ dialog to edit docType schema """
  def __init__(self, comm:Communicate):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
      doc (dict):  document to change / create
    """
    super().__init__()
    self.comm = comm
    self.db   = self.comm.backend.db
    self.docType = ''
    self.df = pd.DataFrame()
    self.closeButtons:list[IconButton] = []  #close buttons of tabs

    # GUI elements
    self.setMinimumWidth(965)
    self.setMinimumHeight(500)
    mainL = QVBoxLayout(self)
    Label('Schema-editor for document types', 'h1', mainL)
    Label('Warning: basic verification exists only. Use with care.', 'h2', mainL)
    _, docTypeL = widgetAndLayout('H', mainL, 's')
    self.selectDocType = QComboBox()
    self.selectDocType.currentTextChanged.connect(self.changeDocType)
    docTypeL.addWidget(self.selectDocType)
    docTypeL.addStretch(1)
    IconButton('fa5s.plus',  self, [Command.NEW],  docTypeL, tooltip='New document type')
    IconButton('fa5s.edit',  self, [Command.EDIT], docTypeL, tooltip='Edit document type')
    IconButton('fa5s.trash', self, [Command.DEL],  docTypeL, tooltip='Delete document type')

    # tabs: empty
    self.tabW = QTabWidget(self)
    self.tabW.tabBarDoubleClicked.connect(self.renameTab)
    self.tabW.tabBarClicked.connect(self.createNewTab)
    mainL.addWidget(self.tabW, stretch=10)
    self.nameColumnDelegates     :list[NameColumnDelegate]      = []
    self.mandatoryColumnDelegates:list[MandatoryColumnDelegate] = []
    self.reorderColumnDelegates  :list[ReorderColumnDelegate]   = []
    self.deleteColumnDelegates   :list[DeleteColumnDelegate]    = []
    self.newWidget = QTableWidget()

    #final button box
    mainL.addStretch(1)
    buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Save | QDialogButtonBox.StandardButton.Cancel)
    buttonBox.clicked.connect(self.closeDialog)
    mainL.addWidget(buttonBox)

    #initialize
    self.readDB()
    self.selectDocType.setCurrentText('x0')


  def readDB(self) -> None:
    """ Read information from database and fill main combobox """
    cmd = 'SELECT docTypeSchema.docType, docTypeSchema.class, docTypeSchema.idx, docTypeSchema.name, '\
          'docTypeSchema.unit, docTypeSchema.mandatory, docTypeSchema.list, definitions.long '\
          'FROM docTypeSchema LEFT JOIN definitions ON definitions.key = (docTypeSchema.class || "." || docTypeSchema.name)'
    self.df = pd.read_sql_query(cmd, self.db.connection).fillna('')
    self.df.rename(columns={'long':'description'}, inplace=True)
    self.df['idx'] = self.df['idx'].astype('int')
    self.selectDocType.clear()
    self.selectDocType.addItems(self.db.dataHierarchy('',''))
    return


  def closeDialog(self, btn:TextButton) -> None:
    """
    cancel or save entered data

    Args:
      btn (QButton): save or cancel button
    """
    if btn.text().endswith('Cancel'):
      self.reject()
    elif btn.text().endswith('Save'):
      self.table2schema()
      # verification: uniqueness in names. etc.
      unique =self.df.groupby('docType')['name'].nunique()==self.df.groupby('docType').size()
      if not unique.all():
        showMessage(self, 'Error', 'Within each table, the text in the first column has to be unique. E.g. no '
                    'two "tags" are allowed.', 'Critical')
        return
      # SAVE DATA
      dfSchema = self.df.drop('description', axis=1)
      dfDef = self.df[['class','name','description']]
      dfDef['key'] = dfDef['class']+'.'+dfDef['name']
      dfDef = dfDef.drop(['name','class'], axis=1)[['key','description']]
      dfDef.rename(columns={'description':'long'}, inplace=True)
      nonUnique = dict(dfDef.groupby(['key']).apply(lambda x: len(np.unique(x))))
      if nonUniqueStr:= ', '.join(k for k,v in nonUnique.items() if v>2):
        button = QMessageBox.question(self, 'Non unique definitions',
                                   f'The definitions are non-unique for {nonUniqueStr}. Do you want to flatten?',
                                   QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
        if button == QMessageBox.StandardButton.No:
          return
      dfDef = dfDef.groupby(['key']).first()
      listDef = list(dfDef.itertuples(name=None))
      # save to sqlite
      dfSchema.to_sql('docTypeSchema', self.db.connection, if_exists='replace', index=False, dtype='str')
      cmd = 'INSERT INTO definitions (key, long, PURL) VALUES (?, ?, "") ON CONFLICT(key) DO '\
            'UPDATE SET long = excluded.long;'
      self.db.cursor.executemany(cmd, listDef)
      self.db.connection.commit()
      self.accept()
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.NEW:
      dialog = DocTypeEditor(self.comm, '', self.changeDocType)
      dialog.exec()
      self.readDB()
    elif command[0] is Command.EDIT:
      dialog = DocTypeEditor(self.comm, self.selectDocType.currentText())
      dialog.exec()
    elif command[0] is Command.DEL:
      button = QMessageBox.question(self, 'Question', 'Do you really want to remove the doc-type?',
                                    QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
      if button == QMessageBox.StandardButton.No:
        return
      self.comm.backend.db.cursor.execute(f"DELETE FROM docTypes WHERE docType == '{self.selectDocType.currentText()}'")
      self.comm.backend.db.cursor.execute(f"DELETE FROM docTypeSchema WHERE docType == '{self.selectDocType.currentText()}'")
      self.comm.backend.db.connection.commit()
    elif command[0] is Command.DEL_GROUP:
      button = QMessageBox.question(self, 'Question', 'Do you really want to remove this group?',
                                    QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
      if button == QMessageBox.StandardButton.Yes:
        self.table2schema()
        group = self.tabW.tabBar().tabText(command[1])
        self.df = self.df[self.df['class']!=group]
        self.redrawTabW()
    return


  def redrawTabW(self) -> None:
    """ Redraw tab-widget: all tables """
    self.tabW.clear()
    self.reorderColumnDelegates.clear()
    df = self.df[self.df['docType']==self.docType]
    for group in df['class'].unique():
      data  = df[df['class']==group]
      table = QTableWidget(len(data)+1, 7)
      table.verticalHeader().hide()
      table.setAlternatingRowColors(True)
      table.setHorizontalHeaderLabels(COLUMN_NAMES)
      for idx, width in enumerate(COLUMN_WIDTH[:-1]):
        table.setColumnWidth(idx, width)
      for _,row in data.iterrows():
        idx = int(row['idx'])
        for col in range(5):
          table.setItem(idx, col, QTableWidgetItem(row[COLUMN_NAMES[col]]))
      #Delegates to give function to row
      self.nameColumnDelegates.append(NameColumnDelegate(self.df[self.df['docType']==self.docType], group))
      self.nameColumnDelegates[-1].add_row_signal.connect(self.addRow)
      table.setItemDelegateForColumn(0, self.nameColumnDelegates[-1])
      self.mandatoryColumnDelegates.append(MandatoryColumnDelegate(self.df[self.df['docType']==self.docType], group))
      table.setItemDelegateForColumn(3, self.mandatoryColumnDelegates[-1])
      self.reorderColumnDelegates.append(ReorderColumnDelegate(self.df[self.df['docType']==self.docType], group))
      self.reorderColumnDelegates[-1].re_order_signal.connect(self.reorderRows)
      table.setItemDelegateForColumn(5, self.reorderColumnDelegates[-1])
      self.deleteColumnDelegates.append(DeleteColumnDelegate(self.df[self.df['docType']==self.docType], group))
      self.deleteColumnDelegates[-1].delete_clicked_signal.connect(self.deleteRow)
      table.setItemDelegateForColumn(6, self.deleteColumnDelegates[-1])
      self.tabW.addTab(table, group or 'default')
    # define close buttons on some of the tabs
    self.closeButtons.clear()
    for idx in range(1, self.tabW.count()):
      self.closeButtons.append(IconButton('fa5s.times', self, [Command.DEL_GROUP,idx], None, 'Delete group'))
      self.tabW.tabBar().setTabButton(idx, QTabBar.ButtonPosition.RightSide, self.closeButtons[-1])
      header = table.horizontalHeader()
      header.setStretchLastSection(True)
    self.tabW.addTab(self.newWidget, '+')
    return


  def changeDocType(self, docType:str) -> None:
    """ Change the document type

    Args:
      docType (str): name of document type
    """
    self.docType = docType
    self.redrawTabW()
    return



  def table2schema(self) -> None:
    """ save text content of tables
    - should be executed at each function that calls redrawTabW
    """
    for tabID in range(self.tabW.count()-1):  # -1: ignore last tab +
      group = self.tabW.tabBar().tabText(tabID)
      group = '' if group=='default' else group
      widget:QTableWidget = self.tabW.widget(tabID) # type: ignore[assignment]
      for row in range(widget.rowCount()-1):  # -1: ignore last row as it is meant for adding
        for col in range(5):
          item = widget.item(row,col)
          self.df.at[row, COLUMN_NAMES[col]] = '' if item is None else item.text()
    return


  def reorderRows(self, row:int) -> None:
    """
    Re-order the data rows: row will be moved up

    Args:
      row (int): row of the data to be shifted up.
    """
    if row>0:
      self.table2schema()
      group = self.tabW.tabBar().tabText(self.tabW.currentIndex())
      group = '' if group=='default' else group
      column= (self.df['docType']==self.docType) & (self.df['class']==group)
      self.df.loc[column & (self.df['idx']==row),   'idx']=-1
      self.df.loc[column & (self.df['idx']==row-1), 'idx']=row
      self.df.loc[column & (self.df['idx']==-1),    'idx']=row-1
      self.df = self.df.sort_values('idx').reset_index(drop=True)
      self.redrawTabW()
    return


  def deleteRow(self, row:int) -> None:
    """
    Delete the data row

    Args:
      row (int): row will be deleted
    """
    if row>0:
      self.table2schema()
      group = self.tabW.tabBar().tabText(self.tabW.currentIndex())
      group = '' if group=='default' else group
      column= (self.df['docType']==self.docType) & (self.df['class']==group)
      self.df = self.df.drop(self.df[column & (self.df['idx']==row)].index)
      self.df.loc[column & (self.df['idx']>row), 'idx'] -= 1
      self.redrawTabW()
    return


  def addRow(self) -> None:
    """
    Add a data row
    """
    self.table2schema()
    group = self.tabW.tabBar().tabText(self.tabW.currentIndex())
    group = '' if group=='default' else group
    widget:QTableWidget  = self.tabW.currentWidget()   # type: ignore[assignment]
    newIdx = widget.rowCount()-1
    if widget.currentItem() is None:  #user pressed escape
      return
    name   = widget.currentItem().text()
    newRow = {'docType':self.docType, 'class':group, 'idx':newIdx, 'name':name,
              'unit': '', 'mandatory': '', 'list': '', 'description':''}
    self.df = pd.concat([self.df, pd.DataFrame([newRow])], ignore_index=True)
    self.redrawTabW()
    return


  def renameTab(self, idx:int) -> None:
    """ rename tab after double clicking

    Args:
      idx (int): index of the tab
    """
    if 0 < idx <self.tabW.count()-1:
      textOld = self.tabW.tabText(idx)
      textNew, ok = QInputDialog.getText(self, 'Rename group', 'Enter new group name:', text=textOld)
      if ok and textNew.strip():
        self.tabW.setTabText(idx, textNew.strip())
        column= (self.df['docType']==self.docType) & (self.df['class']==textOld)
        self.df.loc[column, 'class'] = textNew
    return


  def createNewTab(self, idx:int) -> None:
    """ Create new group

    Args:
      idx (int): index of the tab
    """
    if idx == self.tabW.count()-1:
      textNew, ok = QInputDialog.getText(self, 'Create group', 'Enter new group name:', text='')
      if ok and textNew.strip():
        newRow = {'docType':self.docType, 'class':textNew.strip(), 'idx':0, 'name':'item',
              'unit': '', 'mandatory': '', 'list': '', 'description':''}
        self.df = pd.concat([self.df, pd.DataFrame([newRow])], ignore_index=True)
        self.redrawTabW()
    return


class Command(Enum):
  """ Commands used in this file """
  NEW          = 1
  DEL          = 2
  EDIT         = 3
  DEL_GROUP    = 4



""" Enumeration for database names. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: database_names.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
from enum import Enum


class DatabaseNames(Enum):
  """Enumeration for database names.

  This class defines a set of constants representing different database names used in the application.
  Each member of the enumeration corresponds to a specific database with an associated integer value.

  Attributes:
      DataverseDatabase: Represents the Dataverse database with a value of 1.
      PastaProjectGroupDatabase: Represents the Pasta Project Group database with a value of 2.
  """
  DataverseDatabase = 1
  PastaProjectGroupDatabase = 2



# Form implementation generated from reading ui file 'config_dialog_base.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PySide6 import QtCore, QtGui, QtWidgets


class Ui_ConfigDialogBase:
  def setupUi(self, ConfigDialogBase):
    ConfigDialogBase.setObjectName('ConfigDialogBase')
    ConfigDialogBase.resize(1070, 192)
    self.gridLayout = QtWidgets.QGridLayout(ConfigDialogBase)
    self.gridLayout.setObjectName('gridLayout')
    self.buttonBox = QtWidgets.QDialogButtonBox(parent=ConfigDialogBase)
    self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
    self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Save)
    self.buttonBox.setObjectName('buttonBox')
    self.gridLayout.addWidget(self.buttonBox, 2, 0, 1, 1)
    self.mainVerticalLayout = QtWidgets.QVBoxLayout()
    self.mainVerticalLayout.setObjectName('mainVerticalLayout')
    self.serverHorizontalLayout = QtWidgets.QHBoxLayout()
    self.serverHorizontalLayout.setObjectName('serverHorizontalLayout')
    self.dataverseServerLabel = QtWidgets.QLabel(parent=ConfigDialogBase)
    self.dataverseServerLabel.setMinimumSize(QtCore.QSize(200, 0))
    self.dataverseServerLabel.setObjectName('dataverseServerLabel')
    self.serverHorizontalLayout.addWidget(self.dataverseServerLabel)
    self.dataverseServerLineEdit = QtWidgets.QLineEdit(parent=ConfigDialogBase)
    self.dataverseServerLineEdit.setClearButtonEnabled(True)
    self.dataverseServerLineEdit.setObjectName('dataverseServerLineEdit')
    self.serverHorizontalLayout.addWidget(self.dataverseServerLineEdit)
    spacerItem = QtWidgets.QSpacerItem(155, 20, QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Minimum)
    self.serverHorizontalLayout.addItem(spacerItem)
    self.mainVerticalLayout.addLayout(self.serverHorizontalLayout)
    spacerItem1 = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
    self.mainVerticalLayout.addItem(spacerItem1)
    self.apiTokenHorizontalLayout = QtWidgets.QHBoxLayout()
    self.apiTokenHorizontalLayout.setObjectName('apiTokenHorizontalLayout')
    self.apiTokenLabel = QtWidgets.QLabel(parent=ConfigDialogBase)
    self.apiTokenLabel.setMinimumSize(QtCore.QSize(200, 0))
    self.apiTokenLabel.setObjectName('apiTokenLabel')
    self.apiTokenHorizontalLayout.addWidget(self.apiTokenLabel)
    self.apiTokenLineEdit = QtWidgets.QLineEdit(parent=ConfigDialogBase)
    self.apiTokenLineEdit.setEchoMode(QtWidgets.QLineEdit.EchoMode.Password)
    self.apiTokenLineEdit.setClearButtonEnabled(True)
    self.apiTokenLineEdit.setObjectName('apiTokenLineEdit')
    self.apiTokenHorizontalLayout.addWidget(self.apiTokenLineEdit)
    self.apiTokenHelpPushButton = QtWidgets.QPushButton(parent=ConfigDialogBase)
    self.apiTokenHelpPushButton.setMinimumSize(QtCore.QSize(150, 0))
    self.apiTokenHelpPushButton.setObjectName('apiTokenHelpPushButton')
    self.apiTokenHorizontalLayout.addWidget(self.apiTokenHelpPushButton)
    self.mainVerticalLayout.addLayout(self.apiTokenHorizontalLayout)
    spacerItem2 = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
    self.mainVerticalLayout.addItem(spacerItem2)
    self.dataverseListHorizontalLayout = QtWidgets.QHBoxLayout()
    self.dataverseListHorizontalLayout.setObjectName('dataverseListHorizontalLayout')
    self.dataverseListLabel = QtWidgets.QLabel(parent=ConfigDialogBase)
    self.dataverseListLabel.setMinimumSize(QtCore.QSize(200, 0))
    self.dataverseListLabel.setObjectName('dataverseListLabel')
    self.dataverseListHorizontalLayout.addWidget(self.dataverseListLabel)
    self.dataverseListComboBox = QtWidgets.QComboBox(parent=ConfigDialogBase)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.dataverseListComboBox.sizePolicy().hasHeightForWidth())
    self.dataverseListComboBox.setSizePolicy(sizePolicy)
    self.dataverseListComboBox.setSizeAdjustPolicy(QtWidgets.QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
    self.dataverseListComboBox.setObjectName('dataverseListComboBox')
    self.dataverseListHorizontalLayout.addWidget(self.dataverseListComboBox)
    self.dataverseLineEdit = QtWidgets.QLineEdit(parent=ConfigDialogBase)
    self.dataverseLineEdit.setClearButtonEnabled(True)
    self.dataverseLineEdit.setObjectName('dataverseLineEdit')
    self.dataverseListHorizontalLayout.addWidget(self.dataverseLineEdit)
    self.dataverseVerifyLoadPushButton = QtWidgets.QPushButton(parent=ConfigDialogBase)
    self.dataverseVerifyLoadPushButton.setMinimumSize(QtCore.QSize(150, 0))
    self.dataverseVerifyLoadPushButton.setObjectName('dataverseVerifyLoadPushButton')
    self.dataverseListHorizontalLayout.addWidget(self.dataverseVerifyLoadPushButton)
    self.mainVerticalLayout.addLayout(self.dataverseListHorizontalLayout)
    spacerItem3 = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
    self.mainVerticalLayout.addItem(spacerItem3)
    self.gridLayout.addLayout(self.mainVerticalLayout, 0, 0, 1, 1)

    self.retranslateUi(ConfigDialogBase)
    self.buttonBox.accepted.connect(ConfigDialogBase.accept) # type: ignore
    self.buttonBox.rejected.connect(ConfigDialogBase.reject) # type: ignore
    QtCore.QMetaObject.connectSlotsByName(ConfigDialogBase)

  def retranslateUi(self, ConfigDialogBase):
    _translate = QtCore.QCoreApplication.translate
    ConfigDialogBase.setWindowTitle(_translate('ConfigDialogBase', 'Dataverse configuration'))
    self.dataverseServerLabel.setText(_translate('ConfigDialogBase', 'Dataverse URL'))
    self.dataverseServerLineEdit.setToolTip(_translate('ConfigDialogBase', 'Enter the dataverse server URL.'))
    self.dataverseServerLineEdit.setPlaceholderText(_translate('ConfigDialogBase', 'Enter the dataverse server URL, e.g. https://data.fz-juelich.de/'))
    self.apiTokenLabel.setText(_translate('ConfigDialogBase', 'API token'))
    self.apiTokenLineEdit.setToolTip(_translate('ConfigDialogBase', 'Enter the API token.'))
    self.apiTokenLineEdit.setPlaceholderText(_translate('ConfigDialogBase', 'Enter the API token, e.g. c6527048-5bdc-48b0-a1d5-ed1b62c4513b'))
    self.apiTokenHelpPushButton.setToolTip(_translate('ConfigDialogBase', 'Navigate to the help page for generating dataverse API token.'))
    self.apiTokenHelpPushButton.setText(_translate('ConfigDialogBase', 'Help'))
    self.dataverseListLabel.setText(_translate('ConfigDialogBase', 'Dataverse list'))
    self.dataverseListComboBox.setToolTip(_translate('ConfigDialogBase', 'Displays the dataverse list from the server.'))
    self.dataverseLineEdit.setToolTip(_translate('ConfigDialogBase', 'Displays the ID of the selected dataverse from the list.'))
    self.dataverseLineEdit.setPlaceholderText(_translate('ConfigDialogBase', 'Selected Dataverse ID'))
    self.dataverseVerifyLoadPushButton.setToolTip(_translate('ConfigDialogBase', 'Verify API credentials and load available dataverse list from the server.'))
    self.dataverseVerifyLoadPushButton.setText(_translate('ConfigDialogBase', 'Verify && Load'))


if __name__ == '__main__':
    import sys
    app = QtWidgets.QApplication(sys.argv)
    ConfigDialogBase = QtWidgets.QDialog()
    ui = Ui_ConfigDialogBase()
    ui.setupUi(ConfigDialogBase)
    ConfigDialogBase.show()
    sys.exit(app.exec())



"""extract data from vendor
- default jpg image
For tutorials, see extractor_csv.py and extractor_png.py
"""
import base64
import copy
import json
import re
from io import BytesIO
import numpy as np
from PIL import Image


def use(filePath, style={'main':''}, saveFileName=None):
  """
  Args:
    filePath (Path): full path file name
    style    (dict): supplied to guide the display / extraction style = recipe
                     main is / separated hierarchical elements parent->child
                     can contain more elements
    saveFileName (string): if given, save the image to this file-name

  Returns:
    dict: containing image, metaVendor, metaUser, recipe
  """
  # Extractor
  image = Image.open(filePath)
  maxSize = 400
  if max(image.size)>maxSize:
    scale = max(image.size)/maxSize
    newSize = (int(image.size[0]/scale), int(image.size[1]/scale))
    image = image.resize(newSize)
  metaVendor = image.info
  # clean meta_data
  if 'exif' in metaVendor:
    metaVendor['exif'] = re.sub(r'[^\x20-\x7F]','', metaVendor['exif'].decode('utf-8', errors='ignore'))
  for k,v in metaVendor.items():
    try:
      _ = json.dumps(v)
    except Exception:
      metaVendor[k] = str(v)
  imgArr = np.array(image)
  metaUser   = {'number pixel': imgArr.size,
                'dimension': imgArr.shape}

  #save to file
  imageData = Image.fromarray(imgArr)
  if saveFileName is not None:
    imageData.save(saveFileName)

  # convert PIL image to base64
  figfile = BytesIO()
  imageData.save(figfile, format='JPEG')
  imageData = base64.b64encode(figfile.getvalue()).decode()
  imageData = f"data:image/jpg;base64,{imageData}"

  # return everything
  return {'image':imageData, 'style': {'main':'measurement/image'}, 'metaVendor':metaVendor, 'metaUser':metaUser}



# Form implementation generated from reading ui file 'main_dialog_base.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PySide6 import QtCore, QtGui, QtWidgets


class Ui_MainDialogBase:
  def setupUi(self, MainDialogBase):
    MainDialogBase.setObjectName('MainDialogBase')
    MainDialogBase.resize(1261, 782)
    self.gridLayout = QtWidgets.QGridLayout(MainDialogBase)
    self.gridLayout.setObjectName('gridLayout')
    self.buttonBox = QtWidgets.QDialogButtonBox(parent=MainDialogBase)
    self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
    self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel)
    self.buttonBox.setObjectName('buttonBox')
    self.gridLayout.addWidget(self.buttonBox, 3, 0, 1, 1)
    self.mainHorizontalLayout = QtWidgets.QHBoxLayout()
    self.mainHorizontalLayout.setObjectName('mainHorizontalLayout')
    self.mainVerticalLayout = QtWidgets.QVBoxLayout()
    self.mainVerticalLayout.setObjectName('mainVerticalLayout')
    self.scrollAreaHorizontalLayout = QtWidgets.QHBoxLayout()
    self.scrollAreaHorizontalLayout.setObjectName('scrollAreaHorizontalLayout')
    self.projectsScrollArea = QtWidgets.QScrollArea(parent=MainDialogBase)
    self.projectsScrollArea.setWidgetResizable(True)
    self.projectsScrollArea.setObjectName('projectsScrollArea')
    self.projectScrollAreaWidgetContents = QtWidgets.QWidget()
    self.projectScrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 1235, 336))
    self.projectScrollAreaWidgetContents.setObjectName('projectScrollAreaWidgetContents')
    self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.projectScrollAreaWidgetContents)
    self.verticalLayout_3.setObjectName('verticalLayout_3')
    self.projectsScrollAreaVerticalLayout = QtWidgets.QVBoxLayout()
    self.projectsScrollAreaVerticalLayout.setObjectName('projectsScrollAreaVerticalLayout')
    self.verticalLayout_3.addLayout(self.projectsScrollAreaVerticalLayout)
    self.projectsScrollArea.setWidget(self.projectScrollAreaWidgetContents)
    self.scrollAreaHorizontalLayout.addWidget(self.projectsScrollArea)
    self.mainVerticalLayout.addLayout(self.scrollAreaHorizontalLayout)
    self.uploadButtonHorizontalLayout = QtWidgets.QHBoxLayout()
    self.uploadButtonHorizontalLayout.setObjectName('uploadButtonHorizontalLayout')
    spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
    self.uploadButtonHorizontalLayout.addItem(spacerItem)
    self.selectAllPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.selectAllPushButton.setMinimumSize(QtCore.QSize(100, 0))
    self.selectAllPushButton.setObjectName('selectAllPushButton')
    self.uploadButtonHorizontalLayout.addWidget(self.selectAllPushButton)
    self.deselectAllPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.deselectAllPushButton.setMinimumSize(QtCore.QSize(100, 0))
    self.deselectAllPushButton.setObjectName('deselectAllPushButton')
    self.uploadButtonHorizontalLayout.addWidget(self.deselectAllPushButton)
    self.configureUploadPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.configureUploadPushButton.setMinimumSize(QtCore.QSize(100, 0))
    self.configureUploadPushButton.setObjectName('configureUploadPushButton')
    self.uploadButtonHorizontalLayout.addWidget(self.configureUploadPushButton)
    self.editFullMetadataPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.editFullMetadataPushButton.setMinimumSize(QtCore.QSize(100, 0))
    self.editFullMetadataPushButton.setObjectName('editFullMetadataPushButton')
    self.uploadButtonHorizontalLayout.addWidget(self.editFullMetadataPushButton)
    self.uploadPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.uploadPushButton.setMinimumSize(QtCore.QSize(100, 0))
    self.uploadPushButton.setObjectName('uploadPushButton')
    self.uploadButtonHorizontalLayout.addWidget(self.uploadPushButton)
    self.mainVerticalLayout.addLayout(self.uploadButtonHorizontalLayout)
    self.scrollAreaVerticalLayout = QtWidgets.QVBoxLayout()
    self.scrollAreaVerticalLayout.setObjectName('scrollAreaVerticalLayout')
    self.uploadQueueScrollArea = QtWidgets.QScrollArea(parent=MainDialogBase)
    self.uploadQueueScrollArea.setWidgetResizable(True)
    self.uploadQueueScrollArea.setObjectName('uploadQueueScrollArea')
    self.scrollAreaWidgetContents = QtWidgets.QWidget()
    self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 1235, 302))
    self.scrollAreaWidgetContents.setObjectName('scrollAreaWidgetContents')
    self.verticalLayout = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
    self.verticalLayout.setObjectName('verticalLayout')
    self.uploadQueueVerticalLayout = QtWidgets.QVBoxLayout()
    self.uploadQueueVerticalLayout.setObjectName('uploadQueueVerticalLayout')
    self.verticalLayout.addLayout(self.uploadQueueVerticalLayout)
    self.uploadQueueScrollArea.setWidget(self.scrollAreaWidgetContents)
    self.scrollAreaVerticalLayout.addWidget(self.uploadQueueScrollArea)
    self.scrollAreaButtonsHorizontalLayout = QtWidgets.QHBoxLayout()
    self.scrollAreaButtonsHorizontalLayout.setObjectName('scrollAreaButtonsHorizontalLayout')
    spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
    self.scrollAreaButtonsHorizontalLayout.addItem(spacerItem1)
    self.clearFinishedPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.clearFinishedPushButton.setMinimumSize(QtCore.QSize(150, 0))
    self.clearFinishedPushButton.setObjectName('clearFinishedPushButton')
    self.scrollAreaButtonsHorizontalLayout.addWidget(self.clearFinishedPushButton)
    self.cancelAllPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.cancelAllPushButton.setMinimumSize(QtCore.QSize(150, 0))
    self.cancelAllPushButton.setObjectName('cancelAllPushButton')
    self.scrollAreaButtonsHorizontalLayout.addWidget(self.cancelAllPushButton)
    self.showCompletedPushButton = QtWidgets.QPushButton(parent=MainDialogBase)
    self.showCompletedPushButton.setMinimumSize(QtCore.QSize(150, 0))
    self.showCompletedPushButton.setObjectName('showCompletedPushButton')
    self.scrollAreaButtonsHorizontalLayout.addWidget(self.showCompletedPushButton)
    self.scrollAreaVerticalLayout.addLayout(self.scrollAreaButtonsHorizontalLayout)
    self.mainVerticalLayout.addLayout(self.scrollAreaVerticalLayout)
    self.mainHorizontalLayout.addLayout(self.mainVerticalLayout)
    self.gridLayout.addLayout(self.mainHorizontalLayout, 0, 0, 1, 1)
    spacerItem2 = QtWidgets.QSpacerItem(20, 5, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
    self.gridLayout.addItem(spacerItem2, 1, 0, 1, 1)

    self.retranslateUi(MainDialogBase)
    self.buttonBox.accepted.connect(MainDialogBase.accept) # type: ignore
    self.buttonBox.rejected.connect(MainDialogBase.reject) # type: ignore
    QtCore.QMetaObject.connectSlotsByName(MainDialogBase)

  def retranslateUi(self, MainDialogBase):
    _translate = QtCore.QCoreApplication.translate
    MainDialogBase.setWindowTitle(_translate('MainDialogBase', 'Dataverse upload'))
    self.selectAllPushButton.setToolTip(_translate('MainDialogBase', 'Select all the projects listed above.'))
    self.selectAllPushButton.setText(_translate('MainDialogBase', 'Select all'))
    self.deselectAllPushButton.setToolTip(_translate('MainDialogBase', 'Deselect all the projects listed above.'))
    self.deselectAllPushButton.setText(_translate('MainDialogBase', 'Deselect all'))
    self.configureUploadPushButton.setToolTip(_translate('MainDialogBase', 'Configure the contents of projects which need to be uploaded to dataverse.'))
    self.configureUploadPushButton.setText(_translate('MainDialogBase', 'Configure'))
    self.editFullMetadataPushButton.setToolTip(_translate('MainDialogBase', 'Click to edit full/minimal list of metadata to be used for dataverse upload.'))
    self.editFullMetadataPushButton.setText(_translate('MainDialogBase', 'Edit metadata'))
    self.uploadPushButton.setToolTip(_translate('MainDialogBase', 'Start uploading the above selected projects to dataverse.'))
    self.uploadPushButton.setText(_translate('MainDialogBase', 'Start upload'))
    self.uploadQueueScrollArea.setToolTip(_translate('MainDialogBase', "<html><head/><body><p><span style=\" font-style:italic;\">Displays the enqueued lists of PASTA projects to be uploaded to dataverse.</span><span style=\" font-style:italic;\">Users can view individual logs for each project upload, cancel each or all of them and also clear all the finished items anytime.</span></p></body></html>"))
    self.clearFinishedPushButton.setToolTip(_translate('MainDialogBase', 'Clear all finished/cancelled/error uploads.'))
    self.clearFinishedPushButton.setText(_translate('MainDialogBase', 'Clear'))
    self.cancelAllPushButton.setToolTip(_translate('MainDialogBase', 'Cancel all the ongoing uploads.'))
    self.cancelAllPushButton.setText(_translate('MainDialogBase', 'Cancel all'))
    self.showCompletedPushButton.setToolTip(_translate('MainDialogBase', 'Show uploaded projects history.'))
    self.showCompletedPushButton.setText(_translate('MainDialogBase', 'Show completed'))


if __name__ == '__main__':
    import sys
    app = QtWidgets.QApplication(sys.argv)
    MainDialogBase = QtWidgets.QDialog()
    ui = Ui_MainDialogBase()
    ui.setupUi(MainDialogBase)
    MainDialogBase.show()
    sys.exit(app.exec())



""" Metadata frame factory. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: metadata_frame_factory.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
from typing import Any
from pasta_eln.GUI.dataverse.controlled_vocab_frame import ControlledVocabFrame
from pasta_eln.GUI.dataverse.data_type_class_names import DataTypeClassName
from pasta_eln.GUI.dataverse.metadata_frame_base import MetadataFrame
from pasta_eln.GUI.dataverse.primitive_compound_frame import PrimitiveCompoundFrame


def make_primitive_compound_frame(metadata_field: dict[str, Any]) -> MetadataFrame:
  """
  Creates a primitive compound frame based on the provided metadata field.

  Explanation:
      This function creates and returns a primitive compound frame using the given metadata field.

  Args:
      metadata_field (dict[str, Any]): The metadata field to be used in creating the frame.

  Returns:
      MetadataFrame: An instance of the primitive compound frame.
  """
  return PrimitiveCompoundFrame(metadata_field)


def make_controlled_vocab_frame(metadata_field: dict[str, Any]) -> MetadataFrame:
  """
  Creates a controlled vocabulary frame based on the provided metadata field.

  Explanation:
      This function generates a controlled vocabulary frame using the given metadata field.

  Args:
      metadata_field (dict[str, Any]): The metadata field to be used in creating the frame.

  Returns:
      MetadataFrame: An instance of the controlled vocabulary frame.
  """
  return ControlledVocabFrame(metadata_field)


class MetadataFrameFactory:
  """
  Creates a new instance of the MetadataFrameFactory class.

  Explanation:
      This method creates a new instance of the MetadataFrameFactory class.

  """

  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Creates a new instance of the MetadataFrameFactory class.

    Explanation:
        This method creates a new instance of the MetadataFrameFactory class.

    Args:
        *_: Variable length argument list.
        **__: Arbitrary keyword arguments.

    Returns:
        Any: The new instance of the MetadataFrameFactory class.

    """
    return super().__new__(cls)

  def __init__(self) -> None:
    """
    Initializes a new instance of the MetadataFrameFactory class.

    Explanation:
        This method initializes a new instance of the MetadataFrameFactory class.

    """
    self.factory_map: dict[DataTypeClassName, Any] = {
      DataTypeClassName.PRIMITIVE: make_primitive_compound_frame,
      DataTypeClassName.COMPOUND: make_primitive_compound_frame,
      DataTypeClassName.CONTROLLED_VOCAB: make_controlled_vocab_frame
    }

  def make_metadata_frame(self, class_name: DataTypeClassName, metadata_field: dict[str, Any]) -> MetadataFrame:
    """
    Creates a metadata frame based on the specified data type class name and metadata field.

    Explanation:
        This method dynamically creates and returns a metadata frame based on the provided data type class name and metadata field.

    Args:
        class_name (DataTypeClassName): The name of the data type class.
        metadata_field (dict[str, Any]): The metadata field to be used in the frame creation.

    Returns:
        MetadataFrame: An instance of the created metadata frame.

    Raises:
        ValueError: If the provided data type class name is invalid.
    """
    if class_name in self.factory_map:
      return self.factory_map[class_name](metadata_field)
    else:
      raise ValueError(f"Invalid data type class name: {class_name}")



""" Represents the ORM model for the data hierarchy table in the database. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: data_hierarchy_definition_orm_model.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

from typing import Optional
from sqlalchemy.orm import Mapped, mapped_column
from pasta_eln.database.models.orm_model_base import OrmModelBase


class DataHierarchyDefinitionOrmModel(OrmModelBase):
  """Represents a data hierarchy definition model in the database.

  This class defines the structure of the data hierarchy definition, including
  its properties and their mappings to the database columns. It provides a method
  to retrieve the names of the table columns for easy reference.
  """
  __tablename__ = 'definitions'
  doc_type: Mapped[str] = mapped_column('docType', primary_key=True)
  doc_class: Mapped[Optional[str]] = mapped_column('class', primary_key=True)
  index: Mapped[Optional[str]] = mapped_column('idx', primary_key=True)
  name: Mapped[Optional[str]]
  query: Mapped[Optional[str]]
  unit: Mapped[Optional[str]]
  IRI: Mapped[Optional[str]]
  mandatory: Mapped[Optional[str]]
  meta_list: Mapped[Optional[str]] = mapped_column('list')

  @classmethod
  def get_table_columns(cls) -> list[str]:
    """Retrieves the names of the table columns for the data hierarchy definition model.

    This method returns a list of strings representing the column names in the
    definitions table, which can be useful for database operations and schema
    inspections.
    Returns:
        list[str]: A list of column names for the data hierarchy definition model.
    """
    return [
      'doc_type',
      'doc_class',
      'index',
      'name',
      'query',
      'unit',
      'IRI',
      'mandatory',
      'meta_list'
    ]



""" New/Edit dialog (dialog is blocking the main-window, as opposed to create a new widget-window)"""
import copy
import json
import logging
import re
from enum import Enum
from pathlib import Path
from typing import Any, Union
from PySide6.QtCore import QSize, Qt, QTimer  # pylint: disable=no-name-in-module
from PySide6.QtGui import QRegularExpressionValidator  # pylint: disable=no-name-in-module
from PySide6.QtWidgets import (QComboBox, QDialog, QHBoxLayout, QLabel, QLineEdit,  # pylint: disable=no-name-in-module
                               QMessageBox, QPlainTextEdit, QScrollArea, QSizePolicy, QSplitter, QTabWidget, QTextEdit,
                               QVBoxLayout, QWidget)
from ..fixedStringsJson import SQLiteTranslationDict, defaultDataHierarchyNode, minimalDocInForm
from ..guiCommunicate import Communicate
from ..guiStyle import (IconButton, Image, Label, ScrollMessageBox, TextButton, showMessage, widgetAndLayout,
                        widgetAndLayoutForm)
from ..miscTools import flatten
from ..sqlite import MAIN_ORDER
from ..textTools.stringChanges import createDirName, markdownEqualizer
from ._contextMenu import CommandMenu, executeContextMenu, initContextMenu


class Form(QDialog):
  """ New/Edit dialog (dialog is blocking the main-window, as opposed to create a new widget-window)"""
  def __init__(self, comm:Communicate, doc:dict[str,Any]):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
      doc (dict):  document to change / create
    """
    super().__init__()
    self.comm = comm
    self.db   = self.comm.backend.db
    self.doc  = copy.deepcopy(doc)
    if '_attachments' in self.doc:
      del self.doc['_attachments']
    self.flagNewDoc = True
    if 'id' in self.doc or '_ids' in self.doc:
      self.flagNewDoc = False
    if self.flagNewDoc:
      self.setWindowTitle('Create new entry')
      self.doc['name'] = ''
    else:
      self.setWindowTitle('Edit information')
    self.skipKeys = ['image','metaVendor','metaUser','shasum','._projectID','._ids','.name']
    self.allHidden = False
    self.doc = minimalDocInForm | self.doc

    # GUI elements
    mainL = QVBoxLayout(self)
    splitter = QSplitter(Qt.Orientation.Horizontal)
    splitter.setHandleWidth(10)
    splitter.setContentsMargins(0,0,0,0)
    mainL.addWidget(splitter, stretch=2)

    # image
    if 'image' in self.doc:
      imageWSA = QScrollArea()
      imageWSA.setWidgetResizable(True)
      imageW, self.imageL = widgetAndLayout('V', splitter)
      width = self.comm.backend.configuration['GUI']['imageSizeDetails'] \
                if hasattr(self.comm.backend, 'configuration') else 300
      Image(self.doc['image'], self.imageL, anyDimension=width)
      if 'id' in self.doc:
        self.docID= doc['id']  #required for hide to work
        imageW.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        imageW.customContextMenuRequested.connect(lambda pos: initContextMenu(self, pos))
      imageWSA.setWidget(imageW)
      splitter.addWidget(imageWSA)
      self.setMinimumWidth(1000)
    else:
      self.setMinimumWidth(600)

    # create data hierarchy node: data structure
    if self.doc['type']==['x2']:
      self.doc['type'] = ['x1']
    if self.doc['type'][0] in self.db.dataHierarchy('', ''):
      rawData = self.db.dataHierarchy(self.doc['type'][0], 'meta')
      dataHierarchyNode = copy.deepcopy([dict(i) for i in rawData])
    else:
      dataHierarchyNode = copy.deepcopy(defaultDataHierarchyNode)
    keysDataHierarchy = [f"{i['class']}.{i['name']}" for i in dataHierarchyNode]
    keysDocOrg = [[str(x) for x in (f'{k}.{k1}' for k1 in self.doc[k])] if isinstance(self.doc[k], dict) else [f'.{k}']
               for k in self.doc if k not in MAIN_ORDER+['branch','qrCodes','tags']]
    for keyInDocNotHierarchy in {i for row in keysDocOrg for i in row}.difference(keysDataHierarchy):
      group = keyInDocNotHierarchy.split('.')[0]
      key = keyInDocNotHierarchy.split('.')[1]
      idx = len([1 for i in dataHierarchyNode if i['class']==group])
      dataHierarchyNode.append({'docType': self.doc['type'][0], 'class':group, 'idx':idx, 'name':key, 'list':''})
    groups = {i['class'] for i in dataHierarchyNode}.difference({'metaVendor','metaUser'})
    # create tabs or not: depending on the number of groups
    self.tabW = QTabWidget() #has count=0 if not connected
    if len(groups)>1:
      self.tabW.setParent(self)
      self.tabW.tabBarClicked.connect(self.changeTabs)
      splitter.addWidget(self.tabW)

    # create forms by looping
    self.formsL = []
    self.allUserElements:list[tuple[str,str]] = []
    for group in groups:
      if len(groups)==1:
        _, formL = widgetAndLayoutForm(splitter, 's')
      else:
        formW, formL = widgetAndLayoutForm(None, 's', top='m')
        self.tabW.addTab(formW, 'Home' if group=='' else group)
      self.formsL.append(formL)
      for name in [i['name'] for i in dataHierarchyNode if i['class']==group]:
        key = f"{group}.{name}"
        defaultValue = self.doc['qrCodes'] if key=='.qrCodes' and 'qrCodes' in self.doc else \
                       self.doc.get(group, {}).get(name, ('','','','')) #tags, name, comment are handled separately
        elementName = f"key_{len(self.allUserElements)}"

        # case list
        if key == '.name' and '_ids' not in self.doc:
          setattr(self, elementName, QLineEdit(self.doc['name']))
          getattr(self, elementName).setStyleSheet(self.comm.palette.get('secondaryText','color'))
          getattr(self, elementName).setValidator(QRegularExpressionValidator('[\\w\\ .-]+'))
          formL.addRow('Name', getattr(self, elementName))
          self.allUserElements.append(('name','LineEdit'))
        elif key == '.tags':
          self.tagsBarMainW, tagsBarMainL = widgetAndLayout('H', spacing='s')
          self.gradeChoices = QComboBox()   #part/combobox that shows grades
          self.gradeChoices.setMaximumWidth(80)
          self.gradeChoices.setIconSize(QSize(0,0))
          self.gradeChoices.addItems(['','\u2605','\u2605'*2,'\u2605'*3,'\u2605'*4,'\u2605'*5])
          gradeTag = [i for i in self.doc['tags'] if i.startswith('_')]
          gradeTagStr = '\u2605'*int(gradeTag[0][1]) if gradeTag else ''
          self.gradeChoices.setCurrentText(gradeTagStr)
          tagsBarMainL.addWidget(self.gradeChoices)
          _, self.tagsBarSubL = widgetAndLayout('H', tagsBarMainL, spacing='s', right='m', left='m') #part which shows all the tags
          self.otherChoices = QComboBox()   #part/combobox that allow user to select
          self.otherChoices.setEditable(True)
          self.otherChoices.setMaximumWidth(100)
          self.otherChoices.setValidator(QRegularExpressionValidator('[a-z]\\w+'))
          self.otherChoices.setIconSize(QSize(0,0))
          self.otherChoices.setInsertPolicy(QComboBox.InsertPolicy.InsertAtBottom)
          tagsBarMainL.addWidget(self.otherChoices)
          formL.addRow(QLabel('Tags:'), self.tagsBarMainW)
          self.allUserElements.append(('tags',''))
          self.updateTagsBar()
          self.otherChoices.currentIndexChanged.connect(self.addTag) #connect to slot only after all painting is done
        elif key in ['.comment', '.content']:
          key = key[1:]
          labelW, labelL = widgetAndLayout('V')
          labelL.addWidget(QLabel(key.capitalize()))
          TextButton('More', self, [Command.FOCUS_AREA, key], labelL, checkable=True)
          rightSideW, rightSideL = widgetAndLayout('V')
          setattr(self, f'buttonBarW_{key}', QWidget())
          getattr(self, f'buttonBarW_{key}').hide()
          buttonBarL = QHBoxLayout(getattr(self, f'buttonBarW_{key}'))
          for name, tooltip in [['bold','Bold text'],['italic','Italic text'],['list-ul','Bullet list'],\
                                      ['list-ol','Numbered list']]:
            IconButton(f'fa5s.{name}', self, [Command.BUTTON_BAR, name, key], buttonBarL, tooltip)
          for i in range(1,4):
            icon = f'mdi.format-header-{str(i)}'
            IconButton(icon, self, [Command.BUTTON_BAR, f'heading{str(i)}', key], buttonBarL, f'Heading {str(i)}')
          rightSideL.addWidget(getattr(self, f'buttonBarW_{key}'))
          textStr = self.doc.get(key, '')
          for k,v in SQLiteTranslationDict.items():
            textStr = textStr.replace(v,k)
          setattr(self, f'textEdit_{key}', QPlainTextEdit(textStr))
          getattr(self, f'textEdit_{key}').setAccessibleName(key)
          getattr(self, f'textEdit_{key}').setTabStopDistance(20)
          getattr(self, f'textEdit_{key}').textChanged.connect(self.textChanged)
          setattr(self, f'textShow_{key}', QTextEdit())
          getattr(self, f'textShow_{key}').setMarkdown(markdownEqualizer(self.doc.get(key, '')))
          getattr(self, f'textShow_{key}').setReadOnly(True)
          getattr(self, f'textShow_{key}').hide()
          splitter= QSplitter()
          splitter.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding))
          splitter.addWidget(getattr(self, f'textEdit_{key}'))
          splitter.addWidget(getattr(self, f'textShow_{key}'))
          rightSideL.addWidget(splitter)
          formL.addRow(labelW, rightSideW)
          self.allUserElements.append((key, 'comment'))
        elif key in self.skipKeys:  #skip non desired ones
          continue
        elif isinstance(defaultValue, list):   #list of items, qrCodes in sample
          if len(defaultValue)>0 and isinstance(defaultValue[0], str):
            setattr(self, elementName, QLineEdit(' '.join(defaultValue)))
            self.allUserElements.append((key,'LineEdit'))
            formL.addRow(QLabel(key.capitalize()), getattr(self, elementName))
          else:
            logging.info('Cannot display value of key=%s: %s. Write unknown value for docID=%s',
                         key, str(defaultValue), self.doc['id'])
        elif (isinstance(defaultValue, tuple) and len(defaultValue)==4 and isinstance(defaultValue[0], str)) or \
              isinstance(defaultValue, str):    #string or tuple
          dataHierarchyItem = [i for i in dataHierarchyNode if i['class']==group and f"{i['class']}.{i['name']}"==key]
          if len(dataHierarchyItem)!=1:
            raise ValueError('more than one dataHierarchyItem')
          label = dataHierarchyItem[0]['name'].capitalize()
          if isinstance(defaultValue, str):
            value = defaultValue
          else:  #tuple
            value = defaultValue[0]
            label += '' if defaultValue[1] is None or defaultValue[1]=='' else f' [{defaultValue[1]}]'
            label += '' if defaultValue[3] is None or defaultValue[3]=='' else f'&nbsp;<b><a href="{defaultValue[3]}">&uArr;</a></b>'
          if dataHierarchyItem[0]['list']: #choice dropdown
            setattr(self, elementName, QComboBox())
            if ',' in dataHierarchyItem[0]['list']:                  #dataHierarchy-defined choices
              getattr(self, elementName).addItems(dataHierarchyItem[0]['list'].split(','))
              getattr(self, elementName).setCurrentText(value)
            else:                                                    #choice among docType
              listDocType = dataHierarchyItem[0]['list']
              getattr(self, elementName).addItem('- no link -', userData='')
              for _, line in self.db.getView(f'viewDocType/{listDocType}').iterrows():
                getattr(self, elementName).addItem(line['name'], userData=line['id'])
                if line['id'] == value:
                  getattr(self, elementName).setCurrentIndex(getattr(self, elementName).count()-1)
            self.allUserElements.append((key,'ComboBox'))
          else:                                   #text area
            setattr(self, elementName, QLineEdit(value))
            self.allUserElements.append((key,'LineEdit'))
          getattr(self, elementName).setStyleSheet(self.comm.palette.get('secondaryText','color'))
          formLabelW = QLabel(label)
          formLabelW.setOpenExternalLinks(True)
          formLabelW.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByMouse)
          formL.addRow(formLabelW, getattr(self, elementName))
        else:
          print(f"**WARNING dialogForm: unknown value type. key:{key}, type:{type(defaultValue)}")
        if group in self.doc and name in self.doc[group]:
          del self.doc[group][name]
          if not self.doc[group]:
            del self.doc[group]
      if group == '':
        # individual key-value items
        self.keyValueListW, self.keyValueListL = widgetAndLayoutForm(None, 's')
        self.keyValueListW.hide()
        self.keyValueLabel = QLabel('Key - values')
        self.keyValueLabel.hide()
        formL.addRow(self.keyValueLabel, self.keyValueListW)
        # add extra questions at bottom of form
        allowDocTypeChange = 'id' in self.doc and self.doc['type'][0][0]!='x'
        if '_ids' in self.doc: #if group edit
          allowDocTypeChange = all(docID[0] != 'x' for docID in self.doc['_ids'])
        allowProjectChange = self.doc['type'][0]!='x0'
        if allowProjectChange or allowDocTypeChange:
          formL.addRow(QLabel('Special properties:'), QLabel('') )
          label = '- unassigned -' if self.flagNewDoc else '- no change -'
        # project change
        if allowProjectChange:
          self.projectComboBox = QComboBox()
          self.projectComboBox.setStyleSheet(self.comm.palette.get('secondaryText','color'))
          self.projectComboBox.addItem(label, userData='')
          for _, line in self.db.getView('viewDocType/x0').iterrows():
            # add all projects but the one that is present
            if 'branch' not in self.doc or \
                all( len(branch['stack']) <= 0 or line['id'] != branch['stack'][0] for branch in self.doc['branch']):
              self.projectComboBox.addItem(line['name'], userData=line['id'])
              if self.doc.get('_projectID','') == line['id']:
                self.projectComboBox.setCurrentIndex(self.projectComboBox.count()-1)
          formL.addRow(QLabel('Project'), self.projectComboBox)
        if '_projectID' in self.doc:
          del self.doc['_projectID']
        # docType change
        if allowDocTypeChange: #if not-new and non-folder
          self.docTypeComboBox = QComboBox()
          self.docTypeComboBox.setStyleSheet(self.comm.palette.get('secondaryText','color'))
          self.docTypeComboBox.addItem(label, userData='')
          for key, value in self.db.dataHierarchy('', 'title'):
            if key[0]!='x':
              self.docTypeComboBox.addItem(value, userData=key)
          self.docTypeComboBox.addItem('_UNIDENTIFIED_', userData='-')
          formL.addRow(QLabel('Data type'), self.docTypeComboBox)
    if [i for i in self.doc if i.startswith('_')]:
      logging.error('There should not be "_" in a doc: %s', str(self.doc))
    # final button box
    _, buttonLineL = widgetAndLayout('H', mainL, 'm')
    if 'branch' in self.doc:
      visibilityIcon = all(all(branch['show']) for branch in self.doc['branch'])
      self.visibilityText = QLabel('' if visibilityIcon else 'HIDDEN     \U0001F441')
      buttonLineL.addWidget(self.visibilityText)
    buttonLineL.addStretch(1)
    self.btnAddKWPairs = IconButton('ri.menu-add-fill', self, [Command.FORM_ADD_KV],   buttonLineL,
                                    'Add key-value pair', style='border-width:1')
    if not self.flagNewDoc: #existing dataset
      IconButton('fa5s.poll-h',      self, [Command.FORM_SHOW_DOC], buttonLineL, 'Show all information',
                 style='border-width:1')
      IconButton('fa5s.plus-circle', self, [Command.FORM_SAVE_NEXT], buttonLineL, 'Duplicate data set',
                 style='border-width:1')
    saveBtn = TextButton('Save',             self, [Command.FORM_SAVE],     buttonLineL, 'Save changes')
    saveBtn.setShortcut('Ctrl+Return')
    TextButton('Cancel',           self, [Command.FORM_CANCEL],   buttonLineL, 'Discard changes')
    if self.flagNewDoc: #new dataset
      TextButton('Save && Next', self, [Command.FORM_SAVE_NEXT], buttonLineL, 'Save this and handle next')
    # end of creating form autosave
    if (Path.home()/'.pastaELN.temp').is_file():
      with open(Path.home()/'.pastaELN.temp', encoding='utf-8') as fTemp:
        content = json.loads(fTemp.read())
        if self.doc.get('id', '') in content:
          ret = QMessageBox.information(self, 'Information', 'There is unsaved information from a prematurely '+
                    'closed form. Do you want to restore it?\n If you decline, the unsaved information will be'+
                    ' removed.',
                  QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,
                  QMessageBox.StandardButton.Yes)
          if ret==QMessageBox.StandardButton.Yes:
            subContent = content[self.doc.get('id', '')]
            for key in subContent.keys():
              if key in ('comment', 'content'):
                getattr(self, f'textEdit_{key}').setPlainText(subContent[key])
              elif key in ('tags'):
                self.doc[key] = subContent[key]
                self.updateTagsBar()
              elif isinstance(getattr(self, f'key_{key}'), QLineEdit):
                getattr(self, f'key_{key}').setText(subContent[key])
              # skip QCombobox items since cannot be sure that next from has them and they are easy to recreate
          del content[self.doc.get('id', '')]
      with open(Path.home()/'.pastaELN.temp', 'w', encoding='utf-8') as fTemp:
        fTemp.write(json.dumps(content))
    self.checkThreadTimer = QTimer(self)
    self.checkThreadTimer.setInterval(1*60*1000) #1 min
    self.checkThreadTimer.timeout.connect(self.autosave)
    self.checkThreadTimer.start()


  def autosave(self) -> None:
    """ Autosave comment to file """
    if self.comm.backend.configuration['GUI']['autosave'] == 'No':
      return
    # subContent = {'name':getattr(self, 'key_-name').text().strip(), 'tags':self.doc['tags']}
    # for key in self.allKeys:
    #   if key in ['comment','content']:
    #     subContent[key] = getattr(self, f'textEdit_{key}').toPlainText().strip()
    #   elif key in self.skipKeys or not hasattr(self, f'key_{key}'):
    #     continue
    #   elif isinstance(getattr(self, f'key_{key}'), QLineEdit):
    #     subContent[key] = getattr(self, f'key_{key}').text().strip()
    # skip QCombobox items since cannot be sure that next from has them and they are easy to recreate
    if (Path.home()/'.pastaELN.temp').is_file():
      with open(Path.home()/'.pastaELN.temp', encoding='utf-8') as fTemp:
        content = json.loads(fTemp.read())
    else:
      content = {}
    content[self.doc.get('id', '')] = '' #subContent
    with open(Path.home()/'.pastaELN.temp', 'w', encoding='utf-8') as fTemp:
      fTemp.write(json.dumps(content))
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button

    Args:
      command (list): list of commands
    """
    if isinstance(command[0], CommandMenu):
      if executeContextMenu(self, command):
        self.imageL.itemAt(0).widget().setParent(None)
        width = self.comm.backend.configuration['GUI']['imageSizeDetails'] \
                if hasattr(self.comm.backend, 'configuration') else 300
        if 'image' in self.doc:
          Image(self.doc['image'], self.imageL, anyDimension=width)
        visibilityIcon = all(all(branch['show']) for branch in self.doc['branch'])
        self.visibilityText.setText('' if visibilityIcon else 'HIDDEN     \U0001F441')
    elif command[0] is Command.BUTTON_BAR:
      if command[1]=='bold':
        getattr(self, f'textEdit_{command[2]}').insertPlainText('**TEXT**')
      elif command[1]=='italic':
        getattr(self, f'textEdit_{command[2]}').insertPlainText('*TEXT*')
      elif command[1]=='list-ul':
        getattr(self, f'textEdit_{command[2]}').insertPlainText('\n- item 1\n- item 2')
      elif command[1]=='list-ol':
        getattr(self, f'textEdit_{command[2]}').insertPlainText('\n1. item 1\n1. item 2')
      elif command[1].startswith('heading'):
        getattr(self, f'textEdit_{command[2]}').insertPlainText('#' * int(command[1][-1]) +' Heading\n')
    elif command[0] is Command.FOCUS_AREA:
      unknownWidget = []
      idx = 0 if self.tabW.count()==0 else self.tabW.currentIndex()
      if self.allHidden:  #hide the special buttons and show general form
        getattr(self, f'textShow_{command[1]}').hide()
        getattr(self, f'buttonBarW_{command[1]}').hide()
        for i in range(self.formsL[idx].count()):
          widget = self.formsL[idx].itemAt(i).widget()
          if isinstance(widget, (QLabel, QComboBox, QLineEdit)):
            widget.show()
          else:
            unknownWidget.append(i)
        self.tagsBarMainW.show()
        if command[1]=='content' and len(unknownWidget)==5:  #show / hide label and right-side of non-content and non-comment
          self.formsL[idx].itemAt(unknownWidget[0]).widget().show()
          self.formsL[idx].itemAt(unknownWidget[1]).widget().show()
        if command[1]=='comment' and len(unknownWidget)==5:
          self.formsL[idx].itemAt(unknownWidget[2]).widget().show()
          self.formsL[idx].itemAt(unknownWidget[3]).widget().show()
        if self.keyValueListL.count() == 0:
          self.keyValueLabel.hide()
          self.keyValueListW.hide()
      else:  #show buttons to allow for easy markdown edit; hide general form
        getattr(self, f'textShow_{command[1]}').show()
        getattr(self, f'buttonBarW_{command[1]}').show()
        for i in range(self.formsL[idx].count()):
          widget = self.formsL[idx].itemAt(i).widget()
          if isinstance(widget, (QLabel, QComboBox, QLineEdit)):
            widget.hide()
          else:
            unknownWidget.append(i)
        self.tagsBarMainW.hide()
        if command[1]=='content' and len(unknownWidget)==5:
          self.formsL[idx].itemAt(unknownWidget[0]).widget().hide()
          self.formsL[idx].itemAt(unknownWidget[1]).widget().hide()
        if command[1]=='comment' and len(unknownWidget)==5:
          self.formsL[idx].itemAt(unknownWidget[2]).widget().hide()
          self.formsL[idx].itemAt(unknownWidget[3]).widget().hide()
      self.allHidden = not self.allHidden
    elif command[0] is Command.FORM_CANCEL:
      if self.comm.backend.configuration['GUI']['autosave'] == 'Yes':
        ret = QMessageBox.critical(self, 'Warning', 'You will lose the entered information. Do you want to '+
          'save everything to a temporary location?',
          QMessageBox.StandardButton.Cancel | QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,
          QMessageBox.StandardButton.No)
        if ret==QMessageBox.StandardButton.Cancel:
          return
        if ret==QMessageBox.StandardButton.Yes:
          self.autosave()
      self.checkThreadTimer.stop()
      self.reject()
    elif command[0] in (Command.FORM_SAVE, Command.FORM_SAVE_NEXT):
      self.checkThreadTimer.stop()
      if (Path.home()/'.pastaELN.temp').is_file(): #if there is a temporary file
        with open(Path.home()/'.pastaELN.temp', encoding='utf-8') as fTemp: #open it
          content = json.loads(fTemp.read()) #and read its contents
          if self.doc.get('id', '') in content: #if there is an ID in content -> unique identifier?
            del content[self.doc.get('id', '')]# delete the ID
        with open(Path.home()/'.pastaELN.temp', 'w', encoding='utf-8') as fTemp: #then open the temporary file again
          fTemp.write(json.dumps(content)) #and write all content into the file

      # loop through all the subitems
      if self.gradeChoices.currentText():
        self.doc['tags'] = [i for i in self.doc['tags'] if not i.startswith('_')]
        self.doc['tags'].append(f'_{len(self.gradeChoices.currentText())}')
      for idx, (key, guiType) in enumerate(self.allUserElements):
        elementName = f"key_{idx}"
        valueOld = self.doc.get(key, '')
        if key=='name':
          self.doc['name'] = getattr(self, elementName).text().strip()
          if self.doc['name'] == '':
            showMessage(self, 'Error', 'A created item has to have a valid name')
            return
          if self.doc['type'][0]=='x0':  #prevent project-directory names that are identical
            others = self.comm.backend.db.getView('viewDocType/x0All')
            if 'id' in self.doc:
              others = others[others['id']!=self.doc['id']] # filter data frame by own id
            othersList = [createDirName(str(i),'x0', 0) for i in others['name']] #create names
            while createDirName(self.doc['name'],'x0', 0) in othersList:
              if re.search(r'_\d+$', self.doc['name']) is None:
                self.doc['name'] += '_1'
              else:
                self.doc['name'] = '_'.join(self.doc['name'].split('_')[:-1])+'_'+str(int(self.doc['name'].split('_')[-1])+1)
        elif key in ('tags'):  #tags are already saved
        #     'image', 'metaVendor', 'metaUser' or not hasattr(self, f'key_{key}') and not hasattr(self, f'textEdit_{key}')):
          continue
        elif key == '.qrCodes':
          self.doc['qrCodes'] = getattr(self, elementName).text().strip().split('/')
        elif key in ('comment','content'):
          text = getattr(self, f'textEdit_{key}').toPlainText().strip()
          if '_ids' not in self.doc or text:  #if group edit, text has to have text
            self.doc[key] = text
            if key == 'content' and 'branch' in self.doc:
              for branch in self.doc['branch']:
                if branch['path'] is not None and branch['path'].endswith('.md'):
                  with open(self.comm.backend.basePath/branch['path'], 'w', encoding='utf-8') as fOut:
                    fOut.write(self.doc['content'])
                  logging.debug('Wrote new content to %s',branch['path'])
                elif branch['path'] is not None:
                  showMessage(self, 'Information', 'Did update the database but not the file on harddisk, since PASTA-ELN cannot write this format')
        elif isinstance(valueOld, list):  #items that are comma separated in the text-field
          self.doc[key] = getattr(self, elementName).text().strip().split(' ')
        elif isinstance(valueOld, str):
          if guiType=='ComboBox':
            valueNew = getattr(self, elementName).currentText()
            dataNew  = getattr(self, elementName).currentData()  #if docID is stored in currentData
            if ((dataNew is not None and re.search(r'^[a-z\-]-[a-z0-9]{32}$',dataNew) is not None)
                or dataNew==''):
              self.doc[key] = dataNew
            elif valueNew!='- no link -' or dataNew is None:
              self.doc[key] = valueNew
          else:                          #normal text field
            self.doc[key] = getattr(self, elementName).text().strip()
        elif valueOld is None and key in self.doc:  #important entry, set to empty string
          self.doc[key]=''
          print('Is this really needed?')
        else:
          print('**ERROR dialogForm unknown value type',key, valueOld)
      # new key-value pairs
      keyValueList = [self.keyValueListL.itemAt(i).widget().text() for i in range(self.keyValueListL.count())]# type: ignore[attr-defined]
      keyValueDict = dict(zip(keyValueList[::2],keyValueList[1::2] ))
      keyValueDict = {k:v for k,v in keyValueDict.items() if k}
      self.doc = keyValueDict | self.doc
      # ---- if project changed: only branch save; remaining data still needs saving
      newProjID = self.doc['branch'][0]['stack'] if 'branch' in self.doc else []
      if hasattr(self, 'projectComboBox') and self.projectComboBox.currentData() != '':
        parentPath = self.db.getDoc(self.projectComboBox.currentData())['branch'][0]['path']
        if '_ids' in self.doc:  # group update
          for docID in self.doc['_ids']:
            doc = self.db.getDoc(docID)
            if doc['branch'][0]['stack']!=self.projectComboBox.currentData(): #only if project changed
              if doc['branch'][0]['path'] is None:
                newPath    = ''
              else:
                oldPath    = self.comm.backend.basePath/doc['branch'][0]['path']
                newPath = f'{parentPath}/{oldPath.name}'
                oldPath.rename(self.comm.backend.basePath/newPath)
              self.db.updateBranch( doc['id'], 0, 9999, [self.projectComboBox.currentData()], newPath)
        elif 'branch' in self.doc:             # sequential or single update
          if self.doc['branch'][0]['stack']!=self.projectComboBox.currentData(): #only if project changed
            if self.doc['branch'][0]['path'] is None:
              newPath    = ''
            else:
              oldPath = self.comm.backend.basePath/self.doc['branch'][0]['path']
              newPath = f'{parentPath}/{oldPath.name}'
            self.db.updateBranch( self.doc['id'], 0, 9999, [self.projectComboBox.currentData()], newPath)
            self.doc['branch'][0] = {'stack':[self.projectComboBox.currentData()], 'path':newPath or None, 'child':9999, 'show':[True,True]}
        else:
          newProjID = [self.projectComboBox.currentData()]
      # ---- if docType changed: save; no further save to db required ----
      if hasattr(self, 'docTypeComboBox') and self.docTypeComboBox.currentData() != '':
        self.doc['type'] = [self.docTypeComboBox.currentData()]
      self.doc = flatten(self.doc, True)                  # type: ignore[assignment]
      docBackup = copy.deepcopy(self.doc)                # for duplicate, save&next
      if '_ids' in self.doc:                              # group update
        if 'name' in self.doc:
          del self.doc['name']
        self.doc = {k:v for k,v in self.doc.items() if v} # filter out empty items
        for docID in self.doc.pop('_ids'):
          doc = self.db.getDoc(docID)
          doc.update( self.doc )
          self.comm.backend.editData(doc)
      elif 'id' in self.doc:                             # default update on item
        self.comm.backend.editData(self.doc)
      else:                                               # create new dataset
        self.comm.backend.addData(self.doc['type'][0], copy.deepcopy(self.doc), newProjID)
      self.doc = docBackup
      #!!! NO updates / redraw here since one does not know from where form came
      # e.g. sequential edit cannot have redraw here
      if command[0] is Command.FORM_SAVE_NEXT:
        for delKey in [i for i in self.doc.keys() if i in ['id'] or i.startswith('meta')]: #delete these keys
          del self.doc[delKey]
        self.comm.changeTable.emit('', '')
      else:
        self.accept()  #close
        self.close()
    elif command[0] is Command.FORM_ADD_KV:
      self.keyValueLabel.show()
      self.keyValueListW.show()
      keyLabel = QLineEdit('')
      keyLabel.setPlaceholderText('key')
      keyLabel.setToolTip('Key (leave empty to delete key-value pair)')
      keyLabel.setValidator(QRegularExpressionValidator('[a-zA-Z0-9]\\S+'))
      value = QLineEdit('')
      value.setPlaceholderText('value')
      self.keyValueListL.addRow(keyLabel, value)
    elif command[0] is Command.FORM_SHOW_DOC:
      doc = self.comm.backend.db.getDoc(self.doc['id'])
      if 'image' in doc:
        del doc['image']
      messageWindow = ScrollMessageBox('Details', doc, style='QScrollArea{min-width:600 px; min-height:400px}')
      ret = messageWindow.exec()
    else:
      print('**ERROR dialogForm: unknown Command ', command)
    return


  def textChanged(self) -> None:
    """
    Text changed in editor -> update the display on the right
    """
    key = self.sender().accessibleName()                                                                     # type: ignore
    getattr(self, f'textShow_{key}').setMarkdown(markdownEqualizer(getattr(self, f'textEdit_{key}').toPlainText()))
    return

  def delTag(self, _:str, tag:str) -> None:
    """
    Clicked button to delete tag
    """
    self.doc['tags'].remove(tag)
    self.updateTagsBar()
    return

  def addTag(self, tag:Union[str,int]) -> None:
    """
    Clicked to add tag. Since one needs to use indexChanged to allow the user to enter text, that delivers a int. To allow to differentiate
    between both comboboxes, they cannot be the same (both int), hence grades has to be textChanged

    Args:
      tag (str, int): index (otherTags) or text (grades)
    """
    if isinstance(tag, str):#text from grades
      if tag!='':
        self.doc['tags'] = [i for i in self.doc['tags'] if i[0]!='_']
        self.doc['tags'] += [f'_{len(tag)}']
        self.gradeChoices.setCurrentText('')
    elif tag<1:               #zero index from other-tags
      return
    else:
      tag = self.otherChoices.currentText()
      if tag not in self.doc['tags']:
        self.doc['tags'] += [tag]
      self.otherChoices.setCurrentText('')
    self.updateTagsBar()
    return

  def updateTagsBar(self) -> None:
    """
    After creation, tag removal, tag addition: update the information on screen
    """
    #update tags
    for i in reversed(range(self.tagsBarSubL.count())):
      self.tagsBarSubL.itemAt(i).widget().setParent(None)
    for tag in (self.doc['tags'] if 'tags' in self.doc else []):
      if not re.match(r'^_\d$', tag):
        Label(tag, 'h3', self.tagsBarSubL, self.delTag, tag, 'click to remove')
    self.tagsBarSubL.addWidget(QWidget(), stretch=2)
    #update choices in combobox
    tagsAllList = self.comm.backend.db.getView('viewIdentify/viewTagsAll')['tag'].unique()
    tagsSet = {i for i in tagsAllList if i[0]!='_'}
    newChoicesList = ['']+list(tagsSet.difference([i for i in self.doc['tags'] if i[0]!='_']))
    self.otherChoices.clear()
    self.otherChoices.addItems(newChoicesList)
    return


  def changeTabs(self, idx:int) -> None:
    """
    Clicked on tabs to change to different one

    Args:
      idx (int): index of tab that was clicked
    """
    if idx == 0:
      self.btnAddKWPairs.show()
    else:
      self.btnAddKWPairs.hide()
    return


class Command(Enum):
  """ Commands used in this file """
  BUTTON_BAR       = 1
  CHANGE_EXTRACTOR = 2
  SAVE_IMAGE       = 3
  OPEN_FILEBROWSER = 4
  OPEN_EXTERNAL    = 5
  FOCUS_AREA       = 6
  FORM_SAVE        = 7
  FORM_CANCEL      = 8
  FORM_ADD_KV      = 9
  FORM_SHOW_DOC    = 10
  FORM_SAVE_NEXT   = 11  # same as duplicate



""" Dialog that shows a message and the progress-bar """
import re
from typing import Any, Callable
from PySide6.QtCore import QThread, Signal
from PySide6.QtWidgets import QDialogButtonBox, QProgressBar, QTextBrowser, QVBoxLayout, QWidget


class WaitDialog(QWidget):
  """ Dialog that shows a message and the progress-bar """
  def __init__(self) -> None:
    """ Initialization """
    super().__init__()
    self.count  = 0
    self.mainL = QVBoxLayout()
    self.setMinimumWidth(400)
    self.setMinimumHeight(500)
    self.setWindowTitle('Wait for processes to finish')
    self.setLayout(self.mainL)

    self.text = QTextBrowser()
    self.text.setFixedHeight(450)
    self.text.setMarkdown('Default text')
    self.mainL.addWidget(self.text)
    self.progressBar = QProgressBar(self)
    self.progressBar.setMaximum(100)
    self.progressBar.setValue(0)
    self.mainL.addWidget(self.progressBar)
    self.mainL.addStretch(1)

    #final button box
    self.buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok)
    self.buttonBox.clicked.connect(self.close)
    self.buttonBox.hide()
    self.mainL.addWidget(self.buttonBox)


  def updateProgressBar(self, dType:str, data:str) -> None:
    """ update dialog
    - "text" and "append" will update the text
    - "count" and "incr" will update the progress-bar which runs until 100

    Args:
      dType (str): what to update and how "text", "append", "count", "incr"
      data (str): value to update with
    """
    if dType=='text':
      self.text.setMarkdown(data)
    elif dType=='append':
      self.text.setMarkdown(self.text.toMarkdown().strip()+data)
    elif dType=='incr' and re.match(r'^\d+$',data):
      self.count += int(data)
    elif dType=='count' and re.match(r'^\d+$',data):
      self.count = int(data)
    else:
      print(f"**ERROR unknown data {dType} {data}")
    self.progressBar.setValue(self.count)
    if self.count > 99:
      self.buttonBox.show()
    return



class Worker(QThread):
  """A generic worker thread that runs a given function."""
  progress = Signal(str, str)  # Signal to update the progress bar

  def __init__(self, task_function:Callable[[Callable[[str,str], None]], Any]):
    super().__init__()
    self.task_function = task_function  # Function to execute

  def run(self) -> None:
    """Runs the assigned function, providing a callback for progress updates."""
    try:
      self.task_function(self.progress.emit)  # Pass progress emitter as callback
    except Exception:
      pass
    return



""" Data type class context. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: data_type_class_context.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
from typing import Any
from PySide6.QtWidgets import QFrame, QPushButton, QVBoxLayout


class DataTypeClassContext:
  """
  Context class for a DataTypeClass instance
  """

  def __init__(self,
               main_vertical_layout: QVBoxLayout,
               add_push_button: QPushButton,
               parent_frame: QFrame,
               meta_field: dict[str, Any]) -> None:
    """
    DataTypeClassContext constructor used by DataTypeClassFactory
    Args:
      main_vertical_layout (QVBoxLayout): Main vertical layout in the parent frame
      add_push_button (QPushButton): Add push button in the parent frame
      parent_frame (QFrame): Parent frame
      meta_field (meta_field): Meta field of the data type class
    """
    super().__init__()
    if isinstance(main_vertical_layout, QVBoxLayout):
      self.main_vertical_layout = main_vertical_layout
    else:
      raise ValueError('main_vertical_layout must be a QVBoxLayout!')
    if isinstance(add_push_button, QPushButton):
      self.add_push_button = add_push_button
    else:
      raise ValueError('add_push_button must be a QPushButton!')
    if isinstance(parent_frame, QFrame):
      self.parent_frame = parent_frame
    else:
      raise ValueError('parent_frame must be a QFrame!')
    if isinstance(meta_field, dict):
      self.meta_field = meta_field
    else:
      raise ValueError('meta_field must be a dict!')



""" Contains utility functions. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: utils.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import copy
import re
from asyncio import get_event_loop
from base64 import b64decode, b64encode
from json import dump
from logging import Logger
from os.path import exists, join
from pathlib import Path
from typing import Any, Callable
from cryptography.fernet import Fernet, InvalidToken
from PySide6.QtCore import QSize
from PySide6.QtGui import QImage, QPixmap
from qtawesome import icon
from sqlalchemy import Executable, and_, select
from sqlalchemy.orm import aliased
from pasta_eln.database.models.data_hierarchy_model import DataHierarchyModel
from pasta_eln.database.models.main_orm_model import MainOrmModel
from pasta_eln.database.models.properties_orm_model import PropertiesOrmModel
from pasta_eln.dataverse.client import DataverseClient
from pasta_eln.dataverse.config_error import ConfigError
from pasta_eln.dataverse.pasta_config_reader_factory import PastaConfigReaderFactory
from pasta_eln.dataverse.upload_status_values import UploadStatusValues


def update_status(status: str,
                  status_label_set_text_callback: Callable[[str], None],
                  status_icon_set_pixmap_callback: Callable[[QPixmap | QImage | str], None]) -> None:
  """
  Updates the status with the corresponding label and icon.

  Args:
      status (str): The status value.
      status_label_set_text_callback (Callable[[str], None]): The callback function to set the status label text.
      status_icon_set_pixmap_callback (Callable[[QPixmap | QImage | str], None]): The callback function to set the status icon pixmap.

  Explanation:
      This function updates the status with the corresponding label and icon based on the given status value.
      It uses the provided callback functions to set the status label text and status icon pixmap.
  """
  match status:
    case UploadStatusValues.Queued.name:
      icon_name = 'fa.circle-o-notch'
    case UploadStatusValues.Uploading.name:
      icon_name = 'fa.cloud-upload'
    case UploadStatusValues.Cancelled.name:
      icon_name = 'fa.minus-circle'
    case UploadStatusValues.Finished.name:
      icon_name = 'fa.check-circle-o'
    case UploadStatusValues.Error.name:
      icon_name = 'fa.times-circle-o'
    case UploadStatusValues.Warning.name:
      icon_name = 'fa.warning'
    case _:
      icon_name = 'fa.times-circle-o'
  status_label_set_text_callback(status)
  status_icon_set_pixmap_callback(icon(icon_name).pixmap(QSize(30, 30)))


def set_authors(logger: Logger, metadata: dict[str, Any]) -> None:
  """
  Sets the authors in the metadata.

  Explanation:
      This function sets the authors in the metadata based on the configuration file.
      It retrieves the authors from the configuration file and updates the metadata accordingly.

  Args:
      logger (Logger): The logger instance for logging errors.
      metadata (dict[str, Any]): The metadata dictionary to update.
  """
  config = PastaConfigReaderFactory.get_instance().config
  if config is None:
    raise log_and_create_error(logger, ConfigError, 'Config file not found, Corrupt installation!')
  if 'authors' not in config:
    raise log_and_create_error(logger, ConfigError, 'Incorrect config file, authors not found!')
  author_field = next(
    f for f in metadata['datasetVersion']['metadataBlocks']['citation']['fields'] if f['typeName'] == 'author')
  authors_list = author_field.get('value')
  if authors_list is None:
    raise log_and_create_error(logger, ConfigError, 'Incorrect config file, authors value should be found!')
  template_list = author_field.get('valueTemplate', [{}])
  template = template_list[0] if isinstance(template_list, list) and template_list else {}
  authors_list.clear()
  for author in config['authors']:
    author_copy = copy.deepcopy(template)
    author_copy['authorName']['value'] = ', '.join(filter(None, [author['last'], author['first']]))
    author_copy['authorIdentifierScheme']['value'] = 'ORCID'
    author_copy['authorIdentifier']['value'] = author['orcid']
    author_copy['authorAffiliation']['value'] = ', '.join([o['organization'] for o in author['organizations']])
    authors_list.append(author_copy)


def get_flattened_metadata(metadata: dict[str, Any]) -> dict[str, Any]:
  """
  Returns a flattened version of the metadata.

  Args:
      metadata (dict[str, Any]): The metadata to be flattened.

  Returns:
      dict[str, Any]: The flattened metadata.

  Explanation:
      This function takes a nested metadata dictionary and returns a flattened version of the metadata.
      It extracts the values from the nested structure and creates a new dictionary with a flattened structure.
      The resulting dictionary contains the field names as keys and their corresponding values.

  Exceptions:
      KeyError: If a field name is not found in the metadata dictionary.
  """
  adjusted_metadata = {}
  if metadata['datasetVersion']['license']:
    adjusted_metadata['license'] = metadata['datasetVersion']['license']
  for _, metablock in metadata['datasetVersion']['metadataBlocks'].items():
    for field in metablock['fields']:
      if field['value']:
        adjusted_metadata[field['typeName']] = field['value']
  return adjusted_metadata


def set_template_values(logger: Logger, metadata: dict[str, Any]) -> None:
  """
  Set template values in the metadata.

  This function sets template values in the metadata dictionary based on the type of each field.
  If the metadata is empty, a warning is logged.
  The function iterates through the metadata blocks
  and fields, and for each field, it matches the type class and performs the following actions:
  - For "primitive" or "controlledVocabulary" types:
      - If the field is a multiple type, the value template is set to a copy of the value, and the value is cleared.
      - If the field is a single type, the value template is set to the value, and the value is set to an empty string.
  - For "compound" types:
      - The value template is set to a copy of the value, and the value is cleared.
  - For unknown type classes, a warning is logged.

  Args:
      logger (Logger): The logger object used for logging warnings.
      metadata (dict[str, Any]): The metadata dictionary to update.

  Examples:
      >>> logger = Logger()
      >>> metadata = {'datasetVersion': {'metadataBlocks': {'citation': {'fields': [{'typeClass': 'primitive', 'multiple': False, 'value': 'Example'}]}}}}
      >>> set_template_values(logger, metadata)
      >>> metadata
      {'datasetVersion': {'metadataBlocks': {'citation': {'fields': [{'typeClass': 'primitive', 'multiple': False, 'value': '', 'valueTemplate': 'Example'}]}}}}
  """
  if not metadata:
    logger.warning('Empty metadata, make sure the metadata is loaded correctly...')
    return
  for _, metablock in metadata['datasetVersion']['metadataBlocks'].items():
    for field in metablock['fields']:
      match field['typeClass']:
        case 'primitive':
          set_field_template_value(field)
        case 'controlledVocabulary':
          set_field_template_value(field)
          if field['multiple']:
            field['valueTemplate'].insert(0, 'No Value')
        case 'compound':
          field['valueTemplate'] = field['value'].copy()
          field['value'].clear()
        case _:
          logger.warning(f"Unsupported type class: {field['typeClass']}")


def set_field_template_value(field: dict[str, Any]) -> None:
  """
  Set template values in the field.

  This function sets template values in the field based on the type of the field.
  If the field is a multiple type, the value template is set to a copy of the value, and the value is cleared.
  If the field is a single type, the value template is set to the value, and the value is set to an empty string.

  Args:
      field (dict[str, Any]): The field dictionary to update.

  Examples:
      >>> field = {'typeClass': 'primitive', 'multiple': False, 'value': 'Example'}
      >>> set_field_template_value(field)
      >>> field
      {'typeClass': 'primitive', 'multiple': False, 'value': '', 'valueTemplate': 'Example'}
  """
  if field.get('multiple'):
    if 'value' in field and field['value']:
      field['valueTemplate'] = field['value'].copy()
      field['value'].clear()
    else:
      field['value'] = []
      field['valueTemplate'] = []
  else:
    field['valueTemplate'] = field.get('value')
    if 'value' in field:
      field['value'] = ''


def check_if_minimal_metadata_exists(logger: Logger,
                                     metadata: dict[str, Any],
                                     check_title: bool = True) -> dict[str, list[str]] | None:
  """
  Checks if the minimal metadata exists and returns the missing information.

  Args:
      logger (Logger): The logger object for logging warnings.
      metadata (dict[str, Any]): The metadata dictionary.
      check_title (bool, optional): Flag to indicate if the title should be checked.
      Defaults to True.

  Returns:
      dict[str, list[str]] | None: The missing information dictionary or None if metadata is empty.
      missing_information = {
                              'title': [],
                              'author': [],
                              'datasetContact': [],
                              'dsDescription': [],
                              'subject': []
                            }

  """
  if not metadata:
    logger.warning('Empty metadata, make sure the metadata is loaded correctly...')
    return None
  missing_information: dict[str, list[str]] = {
    'title': [],
    'author': [],
    'datasetContact': [],
    'dsDescription': [],
    'subject': []
  }
  check_if_field_value_not_null(get_citation_field(metadata, 'title'),
                                missing_information,
                                'title',
                                check_title)
  check_if_field_value_not_null(get_citation_field(metadata, 'subject'),
                                missing_information,
                                'subject')
  check_if_compound_field_value_is_missing(
    get_citation_field(metadata, 'author'),
    'author',
    missing_information,
    [('authorName', 'Author Name'),
     ('authorAffiliation', 'Author Affiliation'),
     ('authorIdentifierScheme', 'Author Identifier Scheme'),
     ('authorIdentifier', 'Author Identifier')])
  check_if_compound_field_value_is_missing(
    get_citation_field(metadata, 'datasetContact'),
    'datasetContact',
    missing_information,
    [('datasetContactName', 'Dataset Contact Name'),
     ('datasetContactAffiliation', 'Dataset Contact Affiliation'),
     ('datasetContactEmail', 'Dataset Contact Email')])
  check_if_compound_field_value_is_missing(
    get_citation_field(metadata, 'dsDescription'),
    'dsDescription',
    missing_information,
    [('dsDescriptionValue', 'Dataset Description Value'),
     ('dsDescriptionDate', 'Dataset Description Date')])
  return missing_information


def get_citation_field(metadata: dict[str, Any], name: str) -> dict[str, Any]:
  """
  Retrieves the citation field from the metadata.

  Args:
      metadata (dict[str, Any]): The metadata dictionary.
      name (str): The name of the field to retrieve.

  Returns:
      dict[str, Any]: The citation field.

  """
  return next(
    f for f in metadata['datasetVersion']['metadataBlocks']['citation']['fields'] if f.get('typeName') == name)


def check_if_compound_field_value_is_missing(field: dict[str, Any],
                                             field_key: str,
                                             missing_information: dict[str, list[str]],
                                             sub_fields: list[tuple[str, str]]) -> None:
  """
  Checks if the compound field value is missing and updates the missing_information dictionary if it is.

  Args:
      field (dict[str, Any]): The compound field dictionary.
      field_key (str): The key representing the compound field.
      missing_information (dict[str, list[str]]): The dictionary to track missing information.
      sub_fields (list[tuple[str, str]]): The list of sub-fields to check.
  """
  check_if_field_value_not_null(field, missing_information, field_key)
  if values := field.get('value'):
    for field_value in values:
      for field_name, missing_field_name in sub_fields:
        check_if_field_value_is_missing(field_value, field_key, field_name, missing_field_name,
                                        missing_information)


def check_if_field_value_is_missing(field: dict[str, Any],
                                    field_key: str,
                                    field_name: str,
                                    missing_field_name: str,
                                    missing_information: dict[str, list[str]]) -> None:
  """
  Checks if the field value is missing and updates the missing_information dictionary if it is.

  Args:
      field (dict[str, Any]): The field dictionary.
      field_key (str): The key representing the field.
      field_name (str): The name of the field.
      missing_field_name (str): The name of the missing field.
      missing_information (dict[str, list[str]]): The dictionary to track missing information.

  """
  if not field.get(field_name, {}).get('value'):
    missing_message = f"{missing_field_name} field is missing for one of the {field_key}s!"
    if missing_message not in missing_information[field_key]:
      missing_information[field_key].append(missing_message)


def check_if_field_value_not_null(field: dict[str, Any],
                                  missing_information: dict[str, list[str]],
                                  missing_field_name: str,
                                  check: bool = True) -> None:
  """
  Checks if the field value is not null and updates the missing_information dictionary if it is.

  Args:
      field (dict[str, Any]): The field dictionary.
      missing_information (dict[str, list[str]]): The dictionary to track missing information.
      missing_field_name (str): The name of the missing field.
      check (bool, optional): Flag to indicate if the check should be performed. Defaults to True.

  """
  if check and not field.get('value'):
    missing_message = f"Add at-least a single entry for {missing_field_name.capitalize()}!"
    if missing_message not in missing_information[missing_field_name]:
      missing_information[missing_field_name].append(missing_message)


def get_encrypt_key(logger: Logger) -> tuple[bool, bytes]:
  """
  Gets the dataverse encryption key.

  Explanation:
  This function retrieves the dataverse encryption key from the configuration file. If the key does not exist, a new key is generated and stored in the configuration file.

  Args:
      logger (Logger): The logger instance for logging messages.

  Raises:
    EnvironmentError: If the config file is not found or corrupt.

  Returns:
      tuple[bool, bytes]: A tuple containing a boolean indicating whether the key exists and the encryption key itself.
  """
  logger.info('Getting dataverse encrypt key..')
  config = PastaConfigReaderFactory.get_instance().config
  key_exists = False
  if 'dataverseEncryptKey' not in config:
    logger.warning('Dataverse encrypt key does not exist, hence generating a new key..')
    config['dataverseEncryptKey'] = b64encode(Fernet.generate_key()).decode('ascii')
    write_pasta_config_file(logger, config)
  else:
    key_exists = True
  key = b64decode(config['dataverseEncryptKey'])
  return key_exists, key


def encrypt_data(logger: Logger, encrypt_key: bytes | None, data: str | None) -> str | None:
  """
  Encrypts the provided data using the given encryption key.

  Explanation:
      This function encrypts the provided data using the given encryption key.
      It returns the encrypted data as a string.

  Args:
      logger (Logger): The logger instance for logging information.
      encrypt_key (bytes | None): The encryption key used for encryption.
      data (str | None): The data to be encrypted.

  Returns:
      str | None: The encrypted data as a string, or None if either the encrypt_key or data is None.
  """
  if encrypt_key is None or data is None:
    logger.warning('encrypt_key/data cannot be None')
    return None
  try:
    fernet = Fernet(encrypt_key)
    data = fernet.encrypt(data.encode('ascii')).decode('ascii')
  except InvalidToken as e:
    logger.error('Invalid token: %s', e)
    return None
  except ValueError as e:
    logger.error('Value error: %s', e)
    return None
  except AttributeError as e:
    logger.error('AttributeError: %s', e)
    return None
  return data


def decrypt_data(logger: Logger, encrypt_key: bytes | None, data: str | None) -> str | None:
  """
  Decrypts the provided data using the given encryption key.

  Explanation:
      This function decrypts the provided data using the given encryption key.
      It returns the decrypted data as a string.

  Args:
      logger (Logger): The logger instance for logging information.
      encrypt_key (bytes | None): The encryption key used for decryption.
      data (str | None): The data to be decrypted.

  Returns:
      str | None: The decrypted data as a string, or None if either the encrypt_key or data is None.
  """
  if encrypt_key is None or data is None:
    logger.warning('encrypt_key/data cannot be None')
    return None
  try:
    fernet = Fernet(encrypt_key)
    data = fernet.decrypt(data.encode('ascii')).decode('ascii')
  except InvalidToken as e:
    logger.error('Invalid token: %s', e)
    return None
  except ValueError as e:
    logger.error('Value error: %s', e)
    return None
  except AttributeError as e:
    logger.error('AttributeError: %s', e)
    return None
  return data


def write_pasta_config_file(logger: Logger, config_data: dict[str, Any]) -> None:
  """
  Writes the PASTA config file with the provided configuration data.

  Explanation:
      This function writes the PASTA config file with the provided configuration data.
      It raises a DatabaseError if the config file does not exist.

  Args:
      logger (Logger): The logger instance for logging information.
      config_data (dict[str, Any]): The configuration data to be written to the config file.
  """
  config_file_name = join(Path.home(), '.pastaELN_v3.json')
  if not exists(config_file_name):
    raise log_and_create_error(logger, ConfigError, 'Config file not found, Corrupt installation!')
  logger.info('Writing config file: %s', config_file_name)
  with open(config_file_name, 'w', encoding='utf-8') as conf_file:
    dump(config_data, conf_file, ensure_ascii=False, indent=4)


def log_and_create_error(logger: Logger, exception_type: type[Exception], error_message: str) -> Exception:
  """
  Logs an error message and creates an exception.

  Explanation:
      This function logs the provided error message and creates an exception of the specified type.

  Args:
      logger (Logger): The logger instance for logging errors.
      exception_type (Type[Exception]): The type of exception to create.
      error_message (str): The error message to log and include in the exception.

  Returns:
      Exception: The created exception.
  """
  logger.error(error_message)
  return exception_type(error_message)


def check_login_credentials(logger: Logger, api_token: str, server_url: str) -> tuple[bool, str]:
  """
  Checks if the login credentials (API token and server URL) are valid.

  Explanation:
      This function checks if the provided API token and server URL are valid by
      attempting to reach the dataverse server and verifying the token.

  Args:
      logger (Logger): The logger instance for logging information.
      api_token (str): The API token for authentication.
      server_url (str): The URL of the dataverse server.

  Returns:
      tuple[bool, str]: A tuple containing a boolean indicating if the login credentials are valid
      and a message describing the result.
  """
  logger.info('Checking if login info is valid, server_url: %s', server_url)
  dataverse_client = DataverseClient(server_url, api_token)
  event_loop = get_event_loop()
  success, message = event_loop.run_until_complete(dataverse_client.check_if_dataverse_server_reachable())
  if success:
    token_valid = event_loop.run_until_complete(dataverse_client.check_if_api_token_is_valid())
    if token_valid:
      result = True, 'Data server is reachable and token is valid'
    else:
      result = False, 'Data server is reachable but token is invalid'
      logger.warning('Data server is reachable but API token is invalid!')
  elif 'Unauthorized' in message:
    result = False, 'Data server is reachable but API token is invalid'
    logger.warning('Data server is reachable but API token is invalid: %s', message)
  else:
    result = False, 'Data server is not reachable'
    logger.warning('Data server is not reachable: %s', message)
  return result


def check_if_dataverse_exists(logger: Logger, api_token: str, server_url: str, dataverse_id: str) -> bool:
  """
  Checks if a Dataverse exists based on the provided server URL, API token, and Dataverse ID.

  Args:
      logger (Logger): The logger object for logging information.
      api_token (str): The API token for authentication.
      server_url (str): The URL of the Dataverse server.
      dataverse_id (str): The ID of the Dataverse to check existence for.

  Returns:
      bool: True if the Dataverse exists, False otherwise.
  """
  logger.info('Checking if login info is valid, server_url: %s', server_url)
  dataverse_client = DataverseClient(server_url, api_token)
  event_loop = get_event_loop()
  info = event_loop.run_until_complete(dataverse_client.get_dataverse_info(dataverse_id))
  result = False
  if isinstance(info, dict):
    result = (info.get('id', None) is not None
              and info.get('alias', None) == dataverse_id)
  else:
    logger.warning('Data verse with id %s does not exist, Server message: %s', dataverse_id, info)
  return result


def adjust_type_name(camel_case_string: str) -> str:
  """
  Adjusts the type name from camel case to title case.

  Splits the camel case string into individual words and capitalizes the first letter of each word.
  If a word starts with an uppercase letter, it is left as is.

  Args:
      camel_case_string (str): The input string in camel case.

  Returns:
      str: The adjusted type name in title case.
  """
  split = re.findall(r'[A-Z]?[a-z]+|[A-Z]+(?=[A-Z]|$)',
                     camel_case_string)
  return ' '.join([i.capitalize() if i[0].islower() else i for i in split])


def clear_value(items: dict[str, Any] | None = None) -> None:
  """
  Clears the 'value' attribute of each item in the given dictionary.

  Args:
      items(dict[str, Any] | None): The dictionary containing items to be cleared.

  Explanation:
      This function iterates over the values of the given dictionary and sets the 'value'
      attribute to None for each item.
      If the 'items' argument is None, the function returns immediately.

  """
  if items is None:
    return
  for item in items.values():
    if item and 'value' in item:
      item['value'] = None


def is_date_time_type(type_name: str) -> bool:
  """
  Checks if the given type_name contains 'date' or 'time' as a substring.

  Args:
      type_name (str): The type name to check.

  Returns:
      bool: True if the type_name contains 'date' or 'time', False otherwise.

  Explanation:
      This function checks if the given type_name contains 'date' or 'time' as a substring.
      It performs a case-insensitive search to determine the presence of 'date' or 'time' in the type_name.

  """
  return any(
    map(type_name.lower().__contains__, ['date', 'time']))


def get_formatted_message(missing_metadata: dict[str, list[str]]) -> str:
  """
  Returns a formatted message with missing metadata information.

  Args:
      missing_metadata (dict[str, list[str]]): A dictionary containing missing metadata information.

  Returns:
      str: A formatted message with the missing metadata information.

  Explanation:
      This function takes a dictionary of missing metadata information and returns a formatted message
      with the missing information. The message is formatted as an HTML string and includes a list of missing
      metadata items grouped by their corresponding metadata names.
  """
  if not any(missing_metadata.values()):
    return ''
  name_mapping = {
    'author': 'Author',
    'datasetContact': 'Dataset Contact',
    'dsDescription': 'Dataset Description',
    'subject': 'Subject'
  }
  message = (
    "<html><p><i>Goto 'Edit Metadata' dialog, select 'Minimal' metadata list, enter the below given missing information and retry the upload!</i></p>"
  )
  for metadata_name, missing_list in missing_metadata.items():
    if missing_list:
      message += f"<br></br><b><i>{name_mapping[metadata_name]}:</i></b><ul>"
      for missing in missing_list:
        message += f"<i style=\"color:Crimson\"><li>{missing}</li></i>"
      message += '</ul>'
  message += '</html>'
  return message


def get_formatted_metadata_message(metadata: dict[str, Any]) -> str:
  """
  Formats metadata into an HTML message for display.

  Args:
      metadata (dict): A dictionary containing metadata information.

  Returns:
      str: An HTML-formatted message representing the metadata.

  Examples:
      >>> metadata = {
      ...     'datasetVersion': {
      ...         'license': {'name': 'CC0', 'uri': 'https://creativecommons.org/publicdomain/zero/1.0/'},
      ...         'metadataBlocks': {
      ...             'citation': {
      ...                 'displayName': 'Citation Metadata',
      ...                 'fields': [
      ...                     {'typeName': 'title', 'value': 'Test Title', 'multiple': False, 'typeClass': 'primitive'}
      ...                 ]
      ...             }
      ...         }
      ...     }
      ... }
      >>> get_formatted_metadata_message(metadata)
      '<html>...</html>'
  """

  message = '<html>'
  if metadata['datasetVersion']['license']:
    message += "<b style=\"color:Black\">License Metadata:</b><ul>"
    message += f"<li style=\"color:Gray\">Name: <i>{metadata['datasetVersion']['license']['name']}</i></li>"
    message += f"<li style=\"color:Gray\">URI: <i>{metadata['datasetVersion']['license']['uri']}</i></li>"
    message += '</ul>'
  for _, metablock in metadata['datasetVersion']['metadataBlocks'].items():
    message += f"<b style=\"color:Black\">{metablock['displayName']}:</b><ul>"
    has_value = False
    for field in metablock['fields']:
      if field['value']:
        has_value = has_value or len(field['value']) > 0
        message += f"<b style=\"color:#737373\"><li>{adjust_type_name(field['typeName'])}:</li></b><ul>"
        if field['multiple'] and field['typeClass'] == 'compound':
          for index, field_value in enumerate(field['value']):
            message += f"<li style=\"color:Gray\">Item {index + 1}:</li><ul>"
            for _, value in field_value.items():
              message += f"<li style=\"color:Gray\">{adjust_type_name(value['typeName'])}: <i>{value['value']}</li></i>"
            message += '</ul>'
          message += '</ul>'
        elif not field['multiple'] and field['typeClass'] == 'compound':
          for _, value in field['value'].items():
            message += f"<li style=\"color:Gray\">{adjust_type_name(value['typeName'])}: <i>{value['value']}</li></i>"
          message += '</ul>'
        elif field['multiple'] and field['typeClass'] == 'controlledVocabulary':
          for field_value in field['value']:
            message += f"<i style=\"color:Gray\"><li>{field_value}</li></i>"
          message += '</ul>'
        elif field['multiple'] and field['typeClass'] == 'primitive':
          for field_value in field['value']:
            message += f"<i style=\"color:Gray\"><li>{field_value}</li></i>"
          message += '</ul>'
        else:
          message += f"<i style=\"color:Gray\"><li>{field['value']}</li></i></ul>"
    if not has_value:
      message += "<li style=\"color:#737373\">No Value</li></ul>"
    message += '</ul>'
  message += '</html>'
  return message


def get_formatted_dataverse_url(dataverse_url: str) -> str:
  """
  Formats a Dataverse URL.

  Explanation:
      This function takes a Dataverse URL as input and extracts the persistent ID from it.
      It then formats the URL with the persistent ID in an HTML structure for display.

  Args:
      dataverse_url (str): The Dataverse URL to be formatted.

  Returns:
      str: The formatted Dataverse URL in HTML structure.
  """

  formatted_dataverse_url = ''
  if dataverse_url:
    if 'persistentId=' not in dataverse_url:
      return formatted_dataverse_url
    persistent_id = dataverse_url.split('persistentId=', 1)[1]
    formatted_dataverse_url = (
      f"<html><head/><body><p>Dataverse URL: "
      f"<a href='{dataverse_url}'>"
      f"<span style='font-style:italic; text-decoration: underline; color:#77767b;'>{persistent_id}</span>"
      f"</a></p></body></html>")
  return formatted_dataverse_url


def get_data_hierarchy_types(data_hierarchy: list[DataHierarchyModel] | None) -> list[str | Any]:
  """Retrieves a list of unique data hierarchy types from the provided models.

  This function processes a list of DataHierarchyModel instances and extracts
  their document types, excluding specific types ("x0" and "x1"). It returns
  a list of unique, capitalized document types along with an "Unidentified"
  entry if no valid types are found.

  Args:
      data_hierarchy (list[DataHierarchyModel] | None): A list of data hierarchy models
      to extract types from, or None to return an empty list.

  Returns:
      list[str | Any]: A list of unique document types from the data hierarchy models,
      or an empty list if no models are provided.
  """
  if data_hierarchy is None:
    return []
  data_hierarchy_types = []
  for data_model in data_hierarchy:
    if (data_model and data_model.doc_type
        and data_model.doc_type not in ('x0', 'x1')):
      type_capitalized = data_model.doc_type.capitalize()
      if type_capitalized not in data_hierarchy_types:
        data_hierarchy_types.append(type_capitalized)
  data_hierarchy_types.append('Unidentified')
  return data_hierarchy_types


def get_db_info(logger: Logger) -> dict[str, str]:
  """Retrieves database information from the configuration.

  This function reads the configuration to obtain the database path and name
  associated with the default project group. It raises errors if the configuration
  is missing required fields, ensuring that the application has the necessary
  information to connect to the database.

  Args:
      logger (Logger): The logger instance used for logging errors and information.

  Returns:
      dict[str, str]: A dictionary containing the database path and name.

  Raises:
      ConfigError: If the configuration is missing required fields or contains invalid data.
  """
  config = PastaConfigReaderFactory.get_instance().config
  def_project_group_name = config.get('defaultProjectGroup')
  project_groups = config.get('projectGroups')
  if not def_project_group_name or not project_groups:
    raise log_and_create_error(logger, ConfigError,
                               'Incorrect config file, defaultProjectGroup/projectGroups not found!')
  main_group = project_groups.get(def_project_group_name)
  if not main_group:
    raise log_and_create_error(logger, ConfigError,
                               'Incorrect config file, defaultProjectGroup not found in projectGroups!')
  local_info = main_group.get('local')
  if not local_info:
    raise log_and_create_error(logger, ConfigError,
                               'Incorrect config file, user or password not found!')
  db_path = local_info.get('path')
  db_name = local_info.get('database')
  if db_name and db_path:
    return {
      'database_path': db_path,
      'database_name': db_name
    }
  else:
    raise log_and_create_error(logger, ConfigError,
                               'Incorrect config file, database name/path not found!')


def generate_project_join_statement(model_id: str | None) -> Executable:
  """Generates a SQL join statement for project models.

  This function constructs a SQL query to join the main project model with its
  associated properties based on the provided model ID. It allows for flexible
  querying of project data, including optional filtering based on the model ID.

  Args:
      model_id (str | None): The ID of the project model to filter by. If None,
      the query will not filter by model ID.

  Returns:
      Executable: A SQLAlchemy Executable object representing the constructed
      join statement.

  Raises:
      ValueError: If the model ID is invalid or if there is an error in the
      SQL statement construction.
  """
  properties_objective_aliased = aliased(PropertiesOrmModel)
  properties_status_aliased = aliased(PropertiesOrmModel)
  where_condition = and_(MainOrmModel.type == 'x0',
                         MainOrmModel.id == model_id) if model_id else MainOrmModel.type == 'x0'
  return (select(
    MainOrmModel,
    properties_status_aliased.value,
    properties_objective_aliased.value,
  ).where(where_condition).join_from(
    MainOrmModel,
    properties_objective_aliased,
    and_(
      MainOrmModel.id == properties_objective_aliased.id,
      properties_objective_aliased.key == '.objective',
    ),
    isouter=True,
  ).join_from(
    MainOrmModel,
    properties_status_aliased,
    and_(
      MainOrmModel.id == properties_status_aliased.id,
      properties_status_aliased.key == '.status',
    ),
    isouter=True,
  ))



#!/usr/bin/python3
"""TEST the project view: drag drop randomly items around """
import logging, warnings, random
from pathlib import Path
from PySide6.QtCore import QModelIndex                                  # pylint: disable=no-name-in-module
from pasta_eln.backend import Backend
from pasta_eln.GUI.project import Project
from pasta_eln.guiCommunicate import Communicate
from pasta_eln.GUI.palette import Palette

def test_simple(qtbot):
  """
  main function
  """
  # initialization: create database, destroy on filesystem and database and then create new one
  warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
  warnings.filterwarnings('ignore', message='invalid escape sequence')
  warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
  warnings.filterwarnings('ignore', category=ImportWarning)
  logPath = Path.home()/'pastaELN.log'
  logging.basicConfig(filename=logPath, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                      datefmt='%m-%d %H:%M:%S')   #This logging is always info, since for installation only
  for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
    logging.getLogger(package).setLevel(logging.WARNING)
  logging.info('Start 03 test')

  # start app and load project
  backend = Backend('research')
  palette = Palette(None, 'light_blue')
  comm = Communicate(backend, palette)
  window = Project(comm)
  qtbot.addWidget(window)
  projID = backend.output('x0').split('|')[-1].strip()
  window.change(projID,'')

  choices = random.choices(range(100), k=16)
  # choices = [52,49,38,44,44,91,7,32,55,66,97,34,54,82,84,9]
  print(f'Current choice: [{",".join([str(i) for i in choices])}]')

  # start iteration
  for epoch in range(4):
    print(f'Start drag-drop {epoch}')
    def recursiveRowIteration(index:QModelIndex) -> None:
      item  = window.tree.model().itemFromIndex(index)
      allParentIdx = [index] if item is not None and item.data() is not None and item.data()['docType']==['x1'] else []
      for subRow in range(window.tree.model().rowCount(index)):
        subIndex = window.tree.model().index(subRow,0, index)
        allParentIdx += recursiveRowIteration(subIndex)
      return allParentIdx
    allParentIdx = recursiveRowIteration( window.tree.model().index(-1,0))
    validChoices     = [i for i in allParentIdx if window.tree.model().rowCount(i)>0 ]
    sourceParentIdx  = validChoices[choices.pop(0)%len(validChoices)]
    sourceParentItem = window.tree.model().itemFromIndex(sourceParentIdx)
    validChoices     = range(window.tree.model().rowCount(sourceParentIdx))
    sourceChildRow   = validChoices[choices.pop(0)%len(validChoices)]
    sourceItem       = sourceParentItem.takeChild(sourceChildRow)
    if sourceItem.data() is None:
      continue
    validChoices     = [window.tree.model().itemFromIndex(i) for i in allParentIdx
                        if window.tree.model().itemFromIndex(i).data() is not None and
                          window.tree.model().itemFromIndex(i).data()['hierStack'].split('/')[-1] != sourceItem.data()['hierStack'].split('/')[-1] ]
    targetParent     = validChoices[choices.pop(0)%len(validChoices)]
    print('  ',sourceItem.data(),'->\n  ', targetParent.data())
    validChoices     = range(targetParent.rowCount()+1)
    targetChildRow   = validChoices[choices.pop(0)%len(validChoices)]
    targetParent.setChild(targetChildRow, sourceItem)
    verify(backend)
  return

def verify(backend): # Verify DB
  output = backend.checkDB(outputStyle='text')
  print(output)
  output = '\n'.join(output.split('\n')[8:])
  assert '**ERROR' not in output, 'Error in checkDB'
  return



"""Change the format of dictionaries"""
import difflib
import json
import traceback
import uuid
from datetime import datetime
from typing import Any
from ..fixedStringsJson import SORTED_KEYS, SQLiteTranslation
from .stringChanges import markdownEqualizer


def fillDocBeforeCreate(data:dict[str,Any], docType:list[str]) -> dict[str,Any]:
  """ Fill the data before submission to database with common data
  - type, project, children
  - separate comment into tags, fields
  - create id if needed

  used in backend.py and Store.js

  Args:
    data (dict): document to save
    docType (list): type of document, e.g. project=['x0']

  Returns:
    str: document
  """
  protectedKeys = ['comment','tags','image','externalId']
  # Handle the important entries: type, id, dateCreated, branch, gui
  if 'type' not in data:
    data['type'] = docType
  if data['type']==['']:
    data['type']=['-']
  if isinstance(data['type'], str):
    data['type'] = data['type'].split('/')
  if 'id' not in data:  # if new (if not update): create new id
    data['id'] = data['type'][0][0]+'-'+uuid.uuid4().hex
  if 'externalId' not in data:
    data['externalId'] = ''
  if 'dateCreated' not in data:
    data['dateCreated']   = datetime.now().isoformat()
  data['dateModified']  = datetime.now().isoformat()
  if 'branch' not in data:
    print('Empty branch in data')
    data['branch'] = [{'stack':[], 'path':None, 'child':-1, 'show':[]}]
  if 'show' not in data['branch'] and isinstance(data['branch'], dict):
    data['branch']['show'] = [True]*(len(data['branch']['stack'])+1)
  if 'gui' not in data:
    data['gui'] = [True, True]
  # clean comment and tags
  if 'comment' not in data:
    data['comment'] =''
  if 'tags' not in data:
    data['tags'] = []
  data['tags'] = list(set(data['tags']))  #ensure only one is inside
  #other cleaning
  if ('links' in data and isinstance(data['links'], list)) and \
     (len(data['links'])==0 or (len(data['links'])==1 and data['links'][0]=='')):
    del data['links']
  #individual verification of documents
  if data['type'][0]=='sample':
    if 'qrCodes' not in data:
      data['qrCodes']=[]
    if isinstance(data['qrCodes'], str):
      data['qrCodes'] = data['qrCodes'].split(' ')
  # cleaning at end of all changes
  toDelete = []
  for key in data:
    if isinstance(data[key], str):
      if data[key]=='' and key not in protectedKeys:
        toDelete.append(key)
      else:
        data[key] = data[key].strip()
  for key in toDelete:
    del data[key]
  return data


def dict2ul(aDict:dict[str,Any], fmt:str='html') -> str:
  """
  convert a dictionary into a corresponding <ul>-html list

  Args:
    aDict (dict): dictionary to be translated
    fmt (str): 'html', 'markdown'

  Returns:
    str: string
  """
  text = '<ul>' if fmt=='html' else '\n'
  for key, value in aDict.items():
    if isinstance(value,str) and value.startswith('{') and value.endswith('}'):
      try:
        value = json.loads(value.replace("'",'"'))
      except Exception:
        pass
    if isinstance(value, dict):
      valueString = dict2ul(value)
    elif isinstance(value, tuple) and len(value)==4:  #tuple of value, unit, label, PURL
      key = key if value[2] is None or value[2]=='' else value[2]
      valueString = f'{value[0]} {value[1]}'
      valueString = valueString if value[3] is None or value[3]=='' else f'{valueString}&nbsp;<b><a href="{value[3]}">&uArr;</a></b>'
    elif isinstance(value, (list, tuple)):
      valueString = ',&nbsp;&nbsp;&nbsp;'.join([str(i) for i in value])
    else:
      valueString = str(value)
    text += f'<li>{key}: {valueString}</li>\n' if fmt=='html' else f'- {key}: {valueString}\n'
  return f'{text}</ul>' if fmt=='html' else f'{text}\n'


def doc2markdown(doc:dict[str,Any], ignoreKeys:list[str], dataHierarchyNode:list[dict[str,str]], widget:Any) -> str:
  """
  Convert document to markdown representation

  Args:
    doc (dict): document to be converted
    ignoreKeys (list): keys to be ignored
    dataHierarchyNode (dict): data hierarchy node
    widget (Any): widget to get comm and backend from

  Returns:
    str: markdown representation
  """
  markdown = ''
  for key in [i for i in SORTED_KEYS if i in doc]+[i for i in doc if i not in SORTED_KEYS]:
    value = doc[key]
    if key == '' and isinstance(value,dict):     #handle only key==''
      markdown += doc2markdown(value, ignoreKeys, dataHierarchyNode, widget)
      continue
    if key in ignoreKeys or not value:
      continue
    try:
      if key=='tags':
        tags = ['_curated_' if i=='_curated' else f'#{i}' for i in value]
        tags = ['\u2605'*int(i[2]) if i[:2]=='#_' else i for i in tags]
        markdown += f'Tags: {" ".join(tags)} \n\n'
      elif (isinstance(value,str) and '\n' in value) or key=='comment':                 # long values with /n or comments
        markdown += markdownEqualizer(value)+'\n\n'
      else:
        dataHierarchyItems = [dict(i) for i in dataHierarchyNode if i['name']==key]
        if len(dataHierarchyItems)==1 and 'list' in dataHierarchyItems[0] and dataHierarchyItems[0]['list'] and \
            not isinstance(dataHierarchyItems[0]['list'], list):                #choice among docType
          table  = widget.comm.backend.db.getView('viewDocType/'+dataHierarchyItems[0]['list'])
          names= list(table[table.id==value[0]]['name'])
          if len(names)==1:    # default find one item that we link to
            value = '\u260D '+names[0]
          elif not names:      # likely empty link because the value was not yet defined: just print to show
            value = value[0] if isinstance(value,tuple) else value
          else:
            raise ValueError(f'list target exists multiple times. Key: {key}')
        elif isinstance(value, list):
          value = ', '.join([str(i) for i in value])
          markdown += f'{key.capitalize()}: {value}\n\n'
        if isinstance(value, tuple) and len(value)==4:
          key = key if value[2] is None or value[2]=='' else value[2]
          valueString = f'{value[0]} {value[1]}'
          valueString = valueString if value[3] is None or value[3]=='' else f'{valueString}&nbsp;**[&uArr;]({value[3]})**'
          markdown += f'{key.capitalize()}: {valueString}\n\n'
        if isinstance(value, dict):
          value = dict2ul({k:(v[0]  if isinstance(v, (list,tuple)) else v) for k,v in value.items()}, 'markdown')
          markdown += f'{key.capitalize()}: {value}'
    except Exception:
      doc.pop('image','')
      print(f'**ERROR** Could not convert to markdown value: {value}\n  doc: {doc}\n',traceback.format_exc())
  return markdown


def diffDicts(dict1:dict[str,Any], dict2:dict[str,Any], onlyEssentials:bool=True) -> str:
  """ Check if two dictionaries differ. Just compare the lowest level keys/values and output string

  Args:
    dict1 (dict): dictionary 1 - disk
    dict2 (dict): dictionary 2 - database
    onlyEssentials (bool): compare only the essentials -- see below for includeKeys list

  Returns:
    str: output with \\n
  """
  ignoreKeys = ['client', '_rev', 'gui', '_attachments','externalId','dateSync']
  includeKeys= ['id']
  shortVersion = '__version__' in dict1 and dict1['__version__'] == 'short'
  outString = ''
  dict2Copy = dict(dict2)
  for key,value in dict1.items():
    if key in ignoreKeys or (onlyEssentials and key not in includeKeys):
      continue
    if key not in dict2Copy:
      if not shortVersion or key not in ('__version__','content','image','shasum'):
        print(shortVersion, key)
        outString += f'key not in dictionary 2: {key}' + '\n'
      continue
    if value != dict2Copy[key]:
      if key=='branch':
        if len(value) != len(dict2Copy[key]):
          outString += 'branches have different lengths\n   '+str(value)+'\n   '+str(dict2Copy[key])+'\n'
        else:
          for idx,_ in enumerate(value):
            branch1 = value[idx]
            branch2 = dict2Copy['branch'][idx]
            if 'show' in branch1:
              del branch1['show']
            if 'show' in branch2:
              del branch2['show']
            if branch1!=branch2:
              outString += 'branches differ\n   '+str(value)+'\n   '+str(dict2Copy[key])+'\n'
      elif isinstance(value, list):
        try:
          if set(value).difference(set(dict2Copy[key])):
            outString += (f'lists differ for key: {key}\n   {str(value)}\n   {str(dict2Copy[key])}\n')
        except Exception:
          if str(value)!= str(dict2Copy[key]).replace('(' ,'[').replace(')' ,']'):
            outString += f'Difference in key: {key}\n{str(value)}\n{str(dict2Copy[key])}\n'
      elif isinstance(value, dict):
        if json.dumps(value) != json.dumps(dict2Copy[key]):
          outString += (f'dicts differ for key: {key}\n   {str(value)}\n   {str(dict2Copy[key])}\n')
      elif isinstance(dict2Copy[key], tuple) and len(dict2Copy[key])==4:
        if value != dict2Copy[key][0]:
          outString += (f'property values differ for key: {key}\n   {str(value)}\n   {str(dict2Copy[key])}\n')
      elif isinstance(value,str):
        if value!=dict2Copy[key] and value.translate(SQLiteTranslation)!=dict2Copy[key]:
          diff = difflib.unified_diff(value.splitlines(), dict2Copy[key].splitlines(), fromfile='disk', tofile='database', n=0, lineterm='')
          outString += f'key:{key}\n'+'\n'.join(list(diff))+'\n'
      else:
        outString += (f'values differ for key: {key}\n   {str(value)}\n   {str(dict2Copy[key])}\n')
    del dict2Copy[key]
  extraKeys = set(dict2Copy.keys()).difference(ignoreKeys)
  if shortVersion:
    extraKeys = extraKeys.difference(['tags'])
  if onlyEssentials:
    extraKeys = extraKeys.intersection(includeKeys)
  if extraKeys:
    outString += f'keys not in dictionary 1: {extraKeys}\n'
  return outString


def squashTupleIntoValue(doc:dict[str,Any]) -> None:
  """ Squash tuple/list into value
  - if tuple is length 4, then it is a property tuple: (value, type, unit, description)
  - if tuple is length 1, then it is a value

  Args:
    doc (dict[str,Any]): document
  """
  for key0,value0 in doc.items():
    if isinstance(value0, dict):
      for key1,value1 in doc[key0].items():
        if isinstance(value1, (tuple,list)) and len(value1)==4:
          doc[key0][key1] = value1[0]
  return
